# Copyright 2023 Cisco Systems, Inc. and its affiliates

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, List, cast

from tenacity import retry, retry_if_result, stop_after_attempt, wait_fixed  # type: ignore

from catalystwan.exceptions import TaskValidationError

if TYPE_CHECKING:
    from catalystwan.session import ManagerSession

from catalystwan.endpoints.configuration_dashboard_status import (
    ConfigurationDashboardStatus,
    SubTaskData,
    TaskData,
    TaskResult,
)
from catalystwan.utils.operation_status import OperationStatus, OperationStatusId

logger = logging.getLogger(__name__)


class Task:
    """
    API class for getting data about task/sub-tasks
    """

    def __init__(self, session: ManagerSession, task_id: str):
        self.session = session
        self.task_id = task_id
        self.url = f"/dataservice/device/action/status/{self.task_id}"
        self.task_data: List[SubTaskData]

    def __check_validation_status(self, task: TaskData):
        if not task.validation:
            return None
        if task.validation.status in (OperationStatus.FAILURE, OperationStatus.VALIDATION_FAILURE):
            raise TaskValidationError(f"Task status validation failed, validation status is: {task.validation.status}")

    def wait_for_completed(
        self,
        timeout_seconds: int = 300,
        interval_seconds: int = 5,
        success_statuses: List[OperationStatus] = [
            OperationStatus.SUCCESS,
        ],
        failure_statuses: List[OperationStatus] = [
            OperationStatus.FAILURE,
        ],
        success_statuses_ids: List[OperationStatusId] = [
            OperationStatusId.SUCCESS,
        ],
        failure_statuses_ids: List[OperationStatusId] = [
            OperationStatusId.FAILURE,
        ],
    ) -> TaskResult:
        """
        Method to check subtasks statuses of the task

        Example:
            # create session
            session = create_manager_session(ip_address,admin_username,password,port=port)
            devices = DevicesAPI(session).get()
            vsmart = devices.filter(personality= Personality.VSMART]).single_or_default()

            # Prepare some action, and get it's id
            upgrade_id = SoftwareAction().upgrade_software()

            # Keep asking for reboot status until it's not in exit_statuses (Failure or Success)
            or timeout is not achieved (3000s)
            task = Task(session,upgrade_id).wait_for_completed()
            if task.result:
                #do something
            else:
                #do something else

        Args:


            timeout_seconds (int): After this time, function will stop requesting action status
            interval_seconds (int): interval between action status requests
            validation_timeout_seconds (int): After this time, task validation call will be send
            delay_seconds (int): if Vmanage didn't report task status, after this time api call would be repeated
            success_statuses (Union[List[OperationStatus], str]): list of positive sub-tasks statuses
            success_statuses_ids (Union[List[OperationStatus], str]): list of positive sub-tasks statuses id's
            fails_statuses_id (Union[List[OperationStatusId], str]): list of negative sub-tasks statuses
            fails_statuses_ids (Union[List[OperationStatusId], str]): list of negative sub-tasks statuses id's

        Returns:
            TaskResult(): result attr is True if all subtasks are success
             or is False if at least one is failed
        """
        success_statuses = [cast(OperationStatus, exit_status.value) for exit_status in success_statuses]
        failure_statuses = [cast(OperationStatus, exit_status.value) for exit_status in failure_statuses]
        success_statuses_ids = [
            cast(OperationStatusId, exit_status_id.value) for exit_status_id in success_statuses_ids
        ]
        failure_statuses_ids = [
            cast(OperationStatusId, exit_status_id.value) for exit_status_id in failure_statuses_ids
        ]

        def check_status(task_data: List[SubTaskData]) -> bool:
            """
            Function checks if condition is met. If so,
            wait_for_completed stops asking for task status

            Args:
                task_data (List[SubTaskData]): list of all sub_tasks

            Returns:
                bool: False if condition is met
            """
            if not task_data:
                return True
            task_statuses_success = [task.status in success_statuses for task in task_data]
            task_statuses_failure = [task.status in failure_statuses for task in task_data]
            task_statuses_id_success = [task.status_id in success_statuses_ids for task in task_data]
            task_statuses_id_failure = [task.status_id in failure_statuses_ids for task in task_data]

            all_subtasks_completed_status: List[bool] = [
                any(task_status) for task_status in zip(task_statuses_success, task_statuses_failure)
            ]
            all_subtasks_completed_status_id: List[bool] = [
                any(task_status_id) for task_status_id in zip(task_statuses_id_success, task_statuses_id_failure)
            ]
            if all(all_subtasks_completed_status) or all(all_subtasks_completed_status_id):
                return False
            return True

        def log_exception(self) -> None:
            logger.error("Operation status not achieved in given time")

        @retry(
            wait=wait_fixed(interval_seconds),
            stop=stop_after_attempt(int(timeout_seconds / interval_seconds)),
            retry=retry_if_result(check_status),
            retry_error_callback=log_exception,
        )
        def wait_for_action_finish() -> List[SubTaskData]:
            """
            Keep asking for task status, status_id,
            activity(optional), untill check_status is True

            Returns:
                List[SubTaskData]
            """
            task = ConfigurationDashboardStatus(self.session).find_status(self.task_id)
            self.__check_validation_status(task)
            self.task_data = task.data
            sub_task_statuses = [task.status for task in self.task_data]
            sub_task_statuses_id = [task.status_id for task in self.task_data]
            sub_task_activities = [task.activity for task in self.task_data]
            logger.info(
                f"Sub-tasks data for task {self.task_id}: \n "
                f"statuses: {sub_task_statuses}, status_ids: {sub_task_statuses_id}, activities: {sub_task_activities}."
            )
            return self.task_data

        wait_for_action_finish()
        result = all([sub_task.status in success_statuses for sub_task in self.task_data])
        if result:
            logger.info("Task polling finished, because all subtasks successfully finished.")
        else:
            logger.info("Task polling finished, because at least one subtask failed or task is timeout.")
        return TaskResult(result=result, sub_tasks_data=self.task_data)



# Copyright 2023 Cisco Systems, Inc. and its affiliates

# mypy: disable-error-code="empty-body"
from typing import Dict, List, Optional

from pydantic import BaseModel, ConfigDict, Field

from catalystwan.endpoints import APIEndpoints, delete, get, post, put, versions, view
from catalystwan.models.tenant import Tenant
from catalystwan.typed_list import DataSequence
from catalystwan.utils.session_type import ProviderAsTenantView, ProviderView


class TenantDeleteRequest(BaseModel):
    password: str


class TenantBulkDeleteRequest(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    password: str
    tenant_id_list: List[str] = Field(serialization_alias="tenantIdList", validation_alias="tenantIdList")


class TenantTaskId(BaseModel):
    id: str


class CertificatesStatus(BaseModel):
    invalid: int
    warning: int
    revoked: int


class ControlStatus(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    control_up: int = Field(serialization_alias="controlUp", validation_alias="controlUp")
    partial: int
    control_down: int = Field(serialization_alias="controlDown", validation_alias="controlDown")


class SiteHealth(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    full_connectivity: int = Field(serialization_alias="fullConnectivity", validation_alias="fullConnectivity")
    partial_connectivity: int = Field(serialization_alias="partialConnectivity", validation_alias="partialConnectivity")
    no_connectivity: int = Field(serialization_alias="noConnectivity", validation_alias="noConnectivity")


class vEdgeHealth(BaseModel):
    normal: int
    warning: int
    error: int


class vSmartStatus(BaseModel):
    up: int
    down: int


class TenantStatus(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    tenant_id: str = Field(serialization_alias="tenantId", validation_alias="tenantId")
    tenant_name: str = Field(serialization_alias="tenantName", validation_alias="tenantName")
    control_status: ControlStatus = Field(serialization_alias="controlStatus", validation_alias="controlStatus")
    site_health: SiteHealth = Field(serialization_alias="siteHealth", validation_alias="siteHealth")
    vedge_health: vEdgeHealth = Field(serialization_alias="vEdgeHealth", validation_alias="vEdgeHealth")
    vsmart_status: vSmartStatus = Field(serialization_alias="vSmartStatus", validation_alias="vSmartStatus")


class TenantUpdateRequest(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    tenant_id: str = Field(serialization_alias="tenantId", validation_alias="tenantId")
    subdomain: str = Field(serialization_alias="subDomain", validation_alias="subDomain")
    desc: str
    wan_edge_forecast: Optional[int] = Field(serialization_alias="wanEdgeForecast", validation_alias="wanEdgeForecast")
    edge_connector_enable: Optional[bool] = Field(
        serialization_alias="edgeConnectorEnable", validation_alias="edgeConnectorEnable"
    )
    edge_connector_system_ip: Optional[str] = Field(
        default=None, serialization_alias="edgeConnectorSystemIp", validation_alias="edgeConnectorSystemIp"
    )
    edge_connector_tunnel_interface_name: Optional[str] = Field(
        default=None,
        serialization_alias="edgeConnectorTunnelInterfaceName",
        validation_alias="edgeConnectorTunnelInterfaceName",
    )

    @classmethod
    def from_tenant(cls, tenant: Tenant) -> "TenantUpdateRequest":
        """Creates payload for tenant update from existing tenant data obtained by GET

        Args:
            tenant (Tenant): Tenant to be updated

        Raises:
            TypeError: When provided tenant is missing ID

        Returns:
            TenantUpdateRequest: Tenant attributes suitable for PUT request
        """
        if not tenant.tenant_id:
            raise TypeError("tenantId required for update request")
        return TenantUpdateRequest(
            tenant_id=tenant.tenant_id,
            desc=tenant.desc,
            subdomain=tenant.subdomain,
            wan_edge_forecast=tenant.wan_edge_forecast,
            edge_connector_enable=tenant.edge_connector_enable,
            edge_connector_system_ip=tenant.edge_connector_system_ip,
            edge_connector_tunnel_interface_name=tenant.edge_connector_tunnel_interface_name,
        )


class vSmartPlacementUpdateRequest(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    src_vsmart_uuid: str = Field(serialization_alias="srcvSmartUuid", validation_alias="srcvSmartUuid")
    dest_vsmart_uuid: str = Field(serialization_alias="destvSmartUuid", validation_alias="destvSmartUuid")


class vSmartTenantCapacity(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    vsmart_uuid: str = Field(serialization_alias="vSmartUuid", validation_alias="vSmartUuid")
    total_tenant_capacity: int = Field(
        serialization_alias="totalTenantCapacity", validation_alias="totalTenantCapacity"
    )
    current_tenant_count: int = Field(serialization_alias="currentTenantCount", validation_alias="currentTenantCount")


class vSmartTenantMap(BaseModel):
    data: Dict[str, List[Tenant]]


class vSessionId(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    vsessionid: str = Field(serialization_alias="VSessionId", validation_alias="VSessionId")


class TenantManagement(APIEndpoints):
    @view({ProviderView})
    @post("/tenant")
    def create_tenant(self, payload: Tenant) -> Tenant:
        ...

    @view({ProviderView})
    @post("/tenant/async")
    def create_tenant_async(self, payload: Tenant) -> TenantTaskId:
        ...

    @versions(">=20.4")
    @view({ProviderView})
    @post("/tenant/bulk/async")
    def create_tenant_async_bulk(self, payload: List[Tenant]) -> TenantTaskId:
        ...

    @view({ProviderView})
    @delete("/tenant/{tenant_id}/delete")
    def delete_tenant(self, tenant_id: str, payload: TenantDeleteRequest) -> None:
        ...

    @versions(">=20.4")
    @view({ProviderView})
    @delete("/tenant/bulk/async")
    def delete_tenant_async_bulk(self, payload: TenantBulkDeleteRequest) -> TenantTaskId:
        ...

    def force_status_collection(self):
        # POST /tenantstatus/force
        ...

    @view({ProviderView, ProviderAsTenantView})
    @get("/tenantstatus", "data")
    def get_all_tenant_statuses(self) -> DataSequence[TenantStatus]:
        ...

    @view({ProviderView, ProviderAsTenantView})
    @get("/tenant", "data")
    def get_all_tenants(self) -> DataSequence[Tenant]:
        ...

    @view({ProviderView, ProviderAsTenantView})
    @get("/tenant/{tenant_id}")
    def get_tenant(self, tenant_id: str) -> Tenant:
        ...

    @view({ProviderView})
    @get("/tenant/vsmart/capacity", "data")
    def get_tenant_hosting_capacity_on_vsmarts(self) -> DataSequence[vSmartTenantCapacity]:
        ...

    @view({ProviderView, ProviderAsTenantView})
    @get("/tenant/vsmart")
    def get_tenant_vsmart_mapping(self) -> vSmartTenantMap:
        ...

    def switch_tenant(self):
        # POST /tenant/{tenantId}/switch
        ...

    def tenant_vsmart_mt_migrate(self):
        # POST /tenant/vsmart-mt/migrate
        ...

    @view({ProviderView})
    @put("/tenant/{tenant_id}")
    def update_tenant(self, tenant_id: str, payload: TenantUpdateRequest) -> Tenant:
        ...

    @view({ProviderView})
    @put("/tenant/{tenant_id}/vsmart")
    def update_tenant_vsmart_placement(self, tenant_id: str, payload: vSmartPlacementUpdateRequest) -> None:
        ...

    @view({ProviderView})
    @post("/tenant/{tenant_id}/vsessionid")
    def vsession_id(self, tenant_id: str) -> vSessionId:
        ...



# Copyright 2023 Cisco Systems, Inc. and its affiliates

from ipaddress import IPv4Address, IPv4Network, IPv6Network
from typing import Any, List, Literal, Optional, Set, Tuple
from uuid import UUID

from pydantic import BaseModel, Field

from catalystwan.models.common import InterfaceType, TLOCColor, WellKnownBGPCommunities
from catalystwan.models.policy.lists_entries import (
    AppListEntry,
    AppProbeClassListEntry,
    ASPathListEntry,
    ClassMapListEntry,
    ColorGroupPreference,
    ColorListEntry,
    CommunityListEntry,
    DataIPv6PrefixListEntry,
    DataPrefixListEntry,
    EncapType,
    FQDNListEntry,
    GeoLocationListEntry,
    IPSSignatureListEntry,
    IPv6PrefixListEntry,
    LocalAppListEntry,
    LocalDomainListEntry,
    MirrorListEntry,
    PathPreference,
    PolicerExceedAction,
    PolicerListEntry,
    PortListEntry,
    PreferredColorGroupListEntry,
    PrefixListEntry,
    ProtocolNameListEntry,
    RegionListEntry,
    SiteListEntry,
    SLAClassListEntry,
    TLOCListEntry,
    URLListEntry,
    VPNListEntry,
    ZoneListEntry,
)


class PolicyListBase(BaseModel):
    name: str = Field(
        pattern="^[a-zA-Z0-9_-]{1,32}$",
        description="Can include only alpha-numeric characters, hyphen '-' or underscore '_'; maximum 32 characters",
    )
    description: Optional[str] = "Desc Not Required"
    entries: List[Any]

    def _add_entry(self, entry: Any, single: bool = False) -> None:
        if self.entries and single:
            self.entries[0] = entry
            del self.entries[1:]
        else:
            self.entries.append(entry)


class DataPrefixList(PolicyListBase):
    type: Literal["dataPrefix"] = "dataPrefix"
    entries: List[DataPrefixListEntry] = []

    def add_prefix(self, ip_prefix: IPv4Network) -> None:
        self._add_entry(DataPrefixListEntry(ip_prefix=ip_prefix))


class SiteList(PolicyListBase):
    type: Literal["site"] = "site"
    entries: List[SiteListEntry] = []

    def add_sites(self, sites: Set[int]):
        for site in sites:
            self._add_entry(SiteListEntry(site_id=str(site)))

    def add_site_range(self, site_range: Tuple[int, int]):
        entry = SiteListEntry(site_id=f"{site_range[0]}-{site_range[1]}")
        self._add_entry(entry)


class VPNList(PolicyListBase):
    type: Literal["vpn"] = "vpn"
    entries: List[VPNListEntry] = []

    def add_vpns(self, vpns: Set[int]):
        for vpn in vpns:
            self._add_entry(VPNListEntry(vpn=str(vpn)))

    def add_vpn_range(self, vpn_range: Tuple[int, int]):
        entry = VPNListEntry(vpn=f"{vpn_range[0]}-{vpn_range[1]}")
        self._add_entry(entry)


class ZoneList(PolicyListBase):
    type: Literal["zone"] = "zone"
    entries: List[ZoneListEntry] = []

    def assign_vpns(self, vpns: Set[int]) -> None:
        self.entries = [ZoneListEntry(vpn=str(vpn)) for vpn in vpns]

    def assign_interfaces(self, ifs: Set[InterfaceType]) -> None:
        self.entries = [ZoneListEntry(interface=interface) for interface in ifs]


class FQDNList(PolicyListBase):
    type: Literal["fqdn"] = "fqdn"
    entries: List[FQDNListEntry] = []


class GeoLocationList(PolicyListBase):
    type: Literal["geoLocation"] = "geoLocation"
    entries: List[GeoLocationListEntry] = []


class PortList(PolicyListBase):
    type: Literal["port"] = "port"
    entries: List[PortListEntry] = []


class ProtocolNameList(PolicyListBase):
    type: Literal["protocolName"] = "protocolName"
    entries: List[ProtocolNameListEntry] = []


class LocalAppList(PolicyListBase):
    type: Literal["localApp"] = "localApp"
    entries: List[LocalAppListEntry] = []


class AppList(PolicyListBase):
    type: Literal["app"] = "app"
    entries: List[AppListEntry] = []

    def add_app(self, app: str) -> None:
        self._add_entry(AppListEntry(app=app))

    def add_app_family(self, app_family: str) -> None:
        self._add_entry(AppListEntry(app_family=app_family))


class ColorList(PolicyListBase):
    type: Literal["color"] = "color"
    entries: List[ColorListEntry] = []

    def add_color(self, color: TLOCColor) -> None:
        self._add_entry(ColorListEntry(color=color))


class DataIPv6PrefixList(PolicyListBase):
    type: Literal["dataIpv6Prefix"] = "dataIpv6Prefix"
    entries: List[DataIPv6PrefixListEntry] = []

    def add_prefix(self, ipv6_prefix: IPv6Network) -> None:
        self._add_entry(DataIPv6PrefixListEntry(ipv6_prefix=ipv6_prefix))


class LocalDomainList(PolicyListBase):
    type: Literal["localDomain"] = "localDomain"
    entries: List[LocalDomainListEntry] = []


class IPSSignatureList(PolicyListBase):
    type: Literal["ipsSignature"] = "ipsSignature"
    entries: List[IPSSignatureListEntry] = []


class URLAllowList(PolicyListBase):
    type: Literal["urlWhiteList"] = "urlWhiteList"
    entries: List[URLListEntry] = []


class URLBlockList(PolicyListBase):
    type: Literal["urlBlackList"] = "urlBlackList"
    entries: List[URLListEntry] = []


class _CommunityListBase(PolicyListBase):
    entries: List[CommunityListEntry] = []

    def add_well_known_community(self, community: WellKnownBGPCommunities) -> None:
        self._add_entry(CommunityListEntry(community=community))

    def add_community(self, as_number: int, community_number: int) -> None:
        self._add_entry(CommunityListEntry(community=f"{as_number}:{community_number}"))


class CommunityList(_CommunityListBase):
    type: Literal["community"] = "community"


class ExpandedCommunityList(_CommunityListBase):
    type: Literal["expandedCommunity"] = "expandedCommunity"


class PolicerList(PolicyListBase):
    type: Literal["policer"] = "policer"
    entries: List[PolicerListEntry] = []

    def police(self, burst: int, rate: int, exceed: PolicerExceedAction = "drop") -> None:
        # Policer list must have only single entry!
        entry = PolicerListEntry(burst=str(burst), exceed=exceed, rate=str(rate))
        self._add_entry(entry, single=True)


class ASPathList(PolicyListBase):
    type: Literal["asPath"] = "asPath"
    entries: List[ASPathListEntry] = []


class ClassMapList(PolicyListBase):
    type: Literal["class"] = "class"
    entries: List[ClassMapListEntry] = []

    def assign_queue(self, queue: int) -> None:
        # Class map list must have only one entry!
        entry = ClassMapListEntry(queue=str(queue))
        self._add_entry(entry, single=True)


class MirrorList(PolicyListBase):
    type: Literal["mirror"] = "mirror"
    entries: List[MirrorListEntry] = []


class AppProbeClassList(PolicyListBase):
    type: Literal["appProbe"] = "appProbe"
    entries: List[AppProbeClassListEntry] = []

    def assign_forwarding_class(self, name: str) -> AppProbeClassListEntry:
        # App probe class list must have only one entry!
        entry = AppProbeClassListEntry(forwarding_class=name)
        self._add_entry(entry, single=True)
        return entry


class SLAClassList(PolicyListBase):
    type: Literal["sla"] = "sla"
    entries: List[SLAClassListEntry] = []

    def assign_app_probe_class(
        self,
        app_probe_class_id: UUID,
        latency: Optional[int] = None,
        loss: Optional[int] = None,
        jitter: Optional[int] = None,
    ) -> SLAClassListEntry:
        # SLA class list must have only one entry!
        _latency = str(latency) if latency is not None else None
        _loss = str(loss) if loss is not None else None
        _jitter = str(jitter) if jitter is not None else None
        entry = SLAClassListEntry(latency=_latency, loss=_loss, jitter=_jitter, app_probe_class=app_probe_class_id)
        self._add_entry(entry, single=True)
        return entry

    def add_fallback_jitter_criteria(self, jitter_variance: int) -> None:
        assert self.entries, "Assign app probe class before configuring best fallback tunnel"
        self.entries[0].add_fallback_jitter_criteria(jitter_variance)

    def add_fallback_latency_criteria(self, latency_variance: int) -> None:
        assert self.entries, "Assign app probe class before configuring best fallback tunnel"
        self.entries[0].add_fallback_latency_criteria(latency_variance)

    def add_fallback_loss_criteria(self, loss_variance: int) -> None:
        assert self.entries, "Assign app probe class before configuring best fallback tunnel"
        self.entries[0].add_fallback_loss_criteria(loss_variance)


class TLOCList(PolicyListBase):
    type: Literal["tloc"] = "tloc"
    entries: List[TLOCListEntry] = []

    def add_tloc(self, tloc: IPv4Address, color: TLOCColor, encap: EncapType, preference: Optional[int] = None) -> None:
        _preference = str(preference) if preference is not None else None
        self.entries.append(TLOCListEntry(tloc=tloc, color=color, encap=encap, preference=_preference))


class PreferredColorGroupList(PolicyListBase):
    type: Literal["preferredColorGroup"] = "preferredColorGroup"
    entries: List[PreferredColorGroupListEntry] = []

    def assign_color_groups(
        self,
        primary: Tuple[Set[TLOCColor], PathPreference],
        secondary: Optional[Tuple[Set[TLOCColor], PathPreference]] = None,
        tertiary: Optional[Tuple[Set[TLOCColor], PathPreference]] = None,
    ) -> PreferredColorGroupListEntry:
        primary_preference = ColorGroupPreference.from_color_set_and_path(*primary)
        secondary_preference = (
            ColorGroupPreference.from_color_set_and_path(*secondary) if secondary is not None else None
        )
        tertiary_preference = ColorGroupPreference.from_color_set_and_path(*tertiary) if tertiary is not None else None
        entry = PreferredColorGroupListEntry(
            primary_preference=primary_preference,
            secondary_preference=secondary_preference,
            tertiary_preference=tertiary_preference,
        )
        if self.entries:
            self.entries[0] = entry
        else:
            self.entries.append(entry)
        return entry


class PrefixList(PolicyListBase):
    type: Literal["prefix"] = "prefix"
    entries: List[PrefixListEntry] = []

    def add_prefix(self, prefix: IPv4Network, ge: Optional[int] = None, le: Optional[int] = None) -> None:
        _ge = str(ge) if ge is not None else None
        _le = str(le) if le is not None else None
        self._add_entry(PrefixListEntry(ip_prefix=prefix, ge=_ge, le=_le))


class IPv6PrefixList(PolicyListBase):
    type: Literal["ipv6prefix"] = "ipv6prefix"
    entries: List[IPv6PrefixListEntry] = []


class RegionList(PolicyListBase):
    type: Literal["region"] = "region"
    entries: List[RegionListEntry] = []

    def add_regions(self, regions: Set[int]):
        for region in regions:
            self._add_entry(RegionListEntry(region_id=str(region)))

    def add_region_range(self, region_range: Tuple[int, int]):
        entry = RegionListEntry(region_id=f"{region_range[0]}-{region_range[1]}")
        self._add_entry(entry)



# Copyright 2023 Cisco Systems, Inc. and its affiliates

from enum import Enum, auto


class SessionType(Enum):
    SINGLE_TENANT = auto()
    PROVIDER = auto()
    TENANT = auto()
    PROVIDER_AS_TENANT = auto()
    NOT_DEFINED = auto()


ProviderView = SessionType.PROVIDER
TenantView = SessionType.TENANT
ProviderAsTenantView = SessionType.PROVIDER_AS_TENANT
SingleTenantView = SessionType.SINGLE_TENANT



# Copyright 2023 Cisco Systems, Inc. and its affiliates

from catalystwan.api.templates.models.cisco_system import CiscoSystemModel
from catalystwan.utils.device_model import DeviceModel

default_cisco_system = CiscoSystemModel(
    name="default_cisco_system", description="default", device_models=[DeviceModel.VEDGE_C8000V]
)  # type: ignore



# Copyright 2022 Cisco Systems, Inc. and its affiliates

from enum import Enum


class CertificateStatus(Enum):
    generated = "tokengenerated"
    installed = "certinstalled"


class ValidityPeriod(str, Enum):
    ONE_YEAR = "1Y"
    TWO_YEARS = "2Y"



# Copyright 2023 Cisco Systems, Inc. and its affiliates

from __future__ import annotations

from typing import TYPE_CHECKING, List, Optional

from catalystwan.api.task_status_api import Task
from catalystwan.endpoints.tenant_management import (
    TenantBulkDeleteRequest,
    TenantManagement,
    TenantStatus,
    TenantUpdateRequest,
    vSmartPlacementUpdateRequest,
    vSmartTenantCapacity,
    vSmartTenantMap,
)
from catalystwan.models.tenant import Tenant

if TYPE_CHECKING:
    from catalystwan.session import ManagerSession

from catalystwan.typed_list import DataSequence


class TenantManagementAPI:
    def __init__(self, session: ManagerSession):
        self.session = session
        self._endpoints = TenantManagement(session)

    def get(self) -> DataSequence[Tenant]:
        """Lists all the tenants on the vManage.

        In a multitenant vManage system, this API is only avaiable in the Provider view.

        Returns:
            DataSequence[TenantInfo]: List-like object containing tenant information
        """
        return self._endpoints.get_all_tenants()

    def create(self, tenants: List[Tenant]) -> Task:
        """Creates tenants on vManage

        Args:
            tenants (List[Tenant]): List of tenants to be created

        Returns:
            Task: Object representing tenant creation process
        """
        task_id = self._endpoints.create_tenant_async_bulk(tenants).id
        return Task(self.session, task_id)

    def update(self, tenant_update_request: TenantUpdateRequest) -> Tenant:
        """Updates Tenant on vManage

        Args:
            tenant_update_request (TenantUpdateRequest): Tenant attributes to be updated (must contain tenantId)

        Returns:
            Tenant: Updated tenant data
        """
        return self._endpoints.update_tenant(
            tenant_id=tenant_update_request.tenant_id, tenant_update_request=tenant_update_request
        )

    def delete(self, tenant_id_list: List[str], password: Optional[str] = None) -> Task:
        """Deletes tenants on vManage

        Args:
            tenant_ids (List[str]): Tenant IDs to be deleted
            password (Optional[str]): Provider password if not provided current session password will be used

        Returns:
            Task: Object representing tenant deletion process
        """
        if password is None:
            password = self.session.password
        delete_request = TenantBulkDeleteRequest(tenant_id_list=tenant_id_list, password=password)
        task_id = self._endpoints.delete_tenant_async_bulk(delete_request).id
        return Task(self.session, task_id)

    def get_statuses(self) -> DataSequence[TenantStatus]:
        """Gets tenant statuses from vManage

        Returns:
            DataSequence[TenantStatus]: List-like object containing tenants statuses
        """
        return self._endpoints.get_all_tenant_statuses()

    def get_hosting_capacity_on_vsmarts(self) -> DataSequence[vSmartTenantCapacity]:
        """Gets tenant hosting capacity on vSmarts

        Returns:
            DataSequence[vSmartTenantCapacity]: List-like object containing tenant capacity information for each vSmart
        """
        return self._endpoints.get_tenant_hosting_capacity_on_vsmarts()

    def update_vsmart_placement(self, tenant_id: str, src_vsmart_uuid: str, dst_vsmart_uuid: str):
        """Updates vSmart placement

        Args:
            tenant_id (str): Tenant ID
            src_vsmart_uuid (str): Source vSmart uuid
            dst_vsmart_uuid (str): Destination vSmart uuid
        """
        self._endpoints.update_tenant_vsmart_placement(
            tenant_id=tenant_id,
            vsmart_placement_update_request=vSmartPlacementUpdateRequest(
                src_vsmart_uuid=src_vsmart_uuid, dest_vsmart_uuid=dst_vsmart_uuid
            ),
        )

    def get_vsmart_mapping(self) -> vSmartTenantMap:
        """Gets vSmart to tenant mapping

        Returns:
            vSmartTenantMap: Contains vSmart to tenant mapping
        """
        return self._endpoints.get_tenant_vsmart_mapping()

    def vsession_id(self, tenant_id: str) -> str:
        """Gets VSessionId for given tenant

        Args:
            tenant_id (str): Tenant ID

        Returns:
            str: Contains VSessionId for given tenant
        """
        return self._endpoints.vsession_id(tenant_id).vsessionid



# Copyright 2024 Cisco Systems, Inc. and its affiliates

from pydantic import PlainSerializer
from typing_extensions import Annotated

BoolStr = Annotated[bool, PlainSerializer(lambda x: str(x).lower(), return_type=str, when_used="json-unless-none")]



# Copyright 2024 Cisco Systems, Inc. and its affiliates

from ipaddress import IPv4Address, IPv4Network
from typing import List

from pydantic import AliasPath, BaseModel, ConfigDict, Field

from catalystwan.api.configuration_groups.parcel import Global, _ParcelBase, as_global


class PrefixListEntry(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    ipv4_address: Global[IPv4Address] = Field(serialization_alias="ipv4Address", validation_alias="ipv4Address")
    ipv4_prefix_length: Global[int] = Field(serialization_alias="ipv4PrefixLength", validation_alias="ipv4PrefixLength")


class PrefixListParcel(_ParcelBase):
    entries: List[PrefixListEntry] = Field(default_factory=list, validation_alias=AliasPath("data", "entries"))

    def add_prefix(self, ipv4_network: IPv4Network):
        self.entries.append(
            PrefixListEntry(
                ipv4_address=as_global(ipv4_network.network_address),
                ipv4_prefix_length=as_global(ipv4_network.prefixlen),
            )
        )



# Copyright 2023 Cisco Systems, Inc. and its affiliates

import unittest
from unittest.mock import MagicMock, Mock, patch

from catalystwan.api.partition_manager_api import PartitionManagerAPI
from catalystwan.api.versions_utils import DeviceSoftwareRepository, DeviceVersions, RepositoryAPI
from catalystwan.endpoints.configuration_device_actions import ActionId, RemovePartitionDevice
from catalystwan.endpoints.configuration_device_inventory import DeviceDetailsResponse
from catalystwan.typed_list import DataSequence


class TestPartitionManagerAPI(unittest.TestCase):
    def setUp(self):
        self.device = DataSequence(
            DeviceDetailsResponse,
            [
                DeviceDetailsResponse(
                    personality="vedge",
                    uuid="mock_uuid",
                    device_ip="mock_ip",
                    hostname="mock_host",
                    reachability="reachable",
                    local_system_ip="mock_ip",
                    status="normal",
                    memUsage=1.0,
                    connected_vManages=["192.168.0.1"],
                    model="vedge-cloud",
                )
            ],
        )
        self.DeviceSoftwareRepository_obj = {
            "mock_uuid": DeviceSoftwareRepository(
                installed_versions=["ver1", "ver2", "curr_ver"],
                availableVersions=["ver1", "ver2"],
                version="curr_ver",
                defaultVersion="def_ver",
                uuid="mock_uuid",
            ),
        }

        self.mock_devices = [{"deviceId": "mock_uuid", "deviceIP": "mock_ip", "version": "ver1"}]
        self.mock_device_version_payload = DataSequence(
            RemovePartitionDevice, [RemovePartitionDevice(device_id="mock_uuid", device_ip="mock_ip", version="ver1")]
        )
        mock_session = Mock()
        self.mock_repository_object = RepositoryAPI(mock_session)
        self.mock_device_versions = DeviceVersions(self.mock_repository_object)
        self.mock_partition_manager_obj = PartitionManagerAPI(mock_session)

    @patch("catalystwan.utils.upgrades_helper.get_install_specification")
    @patch.object(DeviceVersions, "get_devices_available_versions")
    def test_remove_partition_if_force_true(self, mock_get_device_list, mock_get_spec):
        # Prepare mock data
        mock_get_device_list.return_value = Mock()
        mock_devices = Mock()
        mock_devices.return_value = self.mock_devices
        mock_get_device_list.return_value = self.mock_device_version_payload
        self.mock_repository_object.devices = self.mock_devices
        expected_id = ActionId(id="mock_action_id")
        self.mock_repository_object.session.endpoints.configuration_device_actions.process_remove_partition = MagicMock(
            return_value=expected_id
        )

        # Assert
        answer = self.mock_partition_manager_obj.remove_partition(self.device, force=True).task_id
        self.assertEqual(answer, "mock_action_id")

    @patch.object(PartitionManagerAPI, "_check_remove_partition_possibility")
    @patch.object(DeviceVersions, "get_devices_available_versions")
    def test_remove_partition_not_raise_error_force_false(self, mock_get_device_list, mock_check_remove):
        # Prepare mock data
        mock_get_device_list.return_value = self.mock_device_version_payload
        mock_check_remove.return_value = []
        self.mock_repository_object.devices = self.mock_devices
        expected_id = ActionId(id="mock_action_id")
        self.mock_repository_object.session.endpoints.configuration_device_actions.process_remove_partition = MagicMock(
            return_value=expected_id
        )

        # Assert
        answer = self.mock_partition_manager_obj.remove_partition(self.device, force=False).task_id
        self.assertEqual(answer, "mock_action_id", "action ids not equal")

    @patch.object(RepositoryAPI, "get_devices_versions_repository")
    def test_check_remove_partition_possibility_if_version_incorrect(self, mock_get_devices_versions_repository):
        # Prepare mock data
        mock_get_devices_versions_repository.return_value = self.DeviceSoftwareRepository_obj
        mock_devices = [
            RemovePartitionDevice(**{"deviceId": "mock_uuid", "deviceIP": "mock_ip", "version": "curr_ver"})
        ]

        # Assert
        with self.assertRaises(ValueError):
            self.mock_partition_manager_obj._check_remove_partition_possibility(mock_devices)

    @patch.object(RepositoryAPI, "get_devices_versions_repository")
    def test_check_remove_partition_possibility_if_version_correct(self, mock_get_devices_versions_repository):
        # Prepare mock data
        mock_get_devices_versions_repository.return_value = self.DeviceSoftwareRepository_obj
        mock_devices = [RemovePartitionDevice(**{"deviceId": "mock_uuid", "deviceIP": "mock_ip", "version": "any_ver"})]

        # Assert
        answer = self.mock_partition_manager_obj._check_remove_partition_possibility(mock_devices)
        self.assertEqual(answer, None, "lists are not equal")

    @patch.object(DeviceVersions, "get_devices_current_version")
    @patch("catalystwan.session.ManagerSession")
    def test_set_default_partition(self, mock_session, mock_get_devices):
        # Arrange
        mock_partition_manager = PartitionManagerAPI(mock_session)
        expected_id = ActionId(id="id_1")
        mock_partition_manager.session.endpoints.configuration_device_actions.process_mark_default_partition = (
            MagicMock(return_value=expected_id)
        )

        # Act
        answer = mock_partition_manager.set_default_partition(self.device).task_id

        # Assert
        self.assertEqual(answer, "id_1")



# Copyright 2023 Cisco Systems, Inc. and its affiliates

from enum import Enum
from pathlib import Path
from typing import ClassVar, List, Optional

from pydantic import ConfigDict, Field

from catalystwan.api.templates.bool_str import BoolStr
from catalystwan.api.templates.feature_template import FeatureTemplate, FeatureTemplateValidator

DEFAULT_BFD_COLOR_MULTIPLIER = 7
DEFAULT_BFD_DSCP = 48
DEFAULT_BFD_HELLO_INTERVAL = 1000
DEFAULT_BFD_POLL_INTERVAL = 600000
DEFAULT_BFD_MULTIPLIER = 6


class ColorType(str, Enum):
    DEFAULT = "default"
    MPLS = "mpls"
    METRO_ETHERNET = "metro-ethernet"
    BIZ_INTERNET = "biz-internet"
    PUBLIC_INTERNET = "public-internet"
    LTE = "lte"
    THREEG = "3g"
    RED = "red"
    GREEN = "green"
    BLUE = "blue"
    GOLD = "gold"
    SILVER = "silver"
    BRONZE = "bronze"
    CUSTOM1 = "custom1"
    CUSTOM2 = "custom2"
    CUSTOM3 = "custom3"
    PRIVATE1 = "private1"
    PRIVATE2 = "private2"
    PRIVATE3 = "private3"
    PRIVATE4 = "private4"
    PRIVATE5 = "private5"
    PRIVATE6 = "private6"


class Color(FeatureTemplateValidator):
    color: ColorType
    hello_interval: Optional[int] = Field(
        DEFAULT_BFD_HELLO_INTERVAL, json_schema_extra={"vmanage_key": "hello-interval"}
    )
    multiplier: Optional[int] = DEFAULT_BFD_COLOR_MULTIPLIER
    pmtu_discovery: Optional[BoolStr] = Field(default=True, json_schema_extra={"vmanage_key": "pmtu-discovery"})
    dscp: Optional[int] = DEFAULT_BFD_DSCP
    model_config = ConfigDict(populate_by_name=True)


class CiscoBFDModel(FeatureTemplate):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    multiplier: Optional[int] = Field(DEFAULT_BFD_MULTIPLIER, json_schema_extra={"data_path": ["app-route"]})
    poll_interval: Optional[int] = Field(
        DEFAULT_BFD_POLL_INTERVAL, json_schema_extra={"vmanage_key": "poll-interval", "data_path": ["app-route"]}
    )
    default_dscp: Optional[int] = Field(DEFAULT_BFD_DSCP, json_schema_extra={"vmanage_key": "default-dscp"})
    color: Optional[List[Color]] = None

    payload_path: ClassVar[Path] = Path(__file__).parent / "DEPRECATED"
    type: ClassVar[str] = "cisco_bfd"



# Copyright 2024 Cisco Systems, Inc. and its affiliates

from typing import List, Literal, Optional, Union
from uuid import UUID

from pydantic import BaseModel, ConfigDict, Field

from catalystwan.api.configuration_groups.parcel import Default, Global, Variable
from catalystwan.models.configuration.feature_profile.sdwan.service.lan.common import (
    Arp,
    StaticIPv4Address,
    StaticIPv6Address,
    VrrpIPv6Address,
    VrrpTrackingObject,
)
from catalystwan.models.configuration.feature_profile.sdwan.service.lan.vpn import Direction

NatType = Literal[
    "pool",
    "loopback",
]

DuplexMode = Literal[
    "full",
    "half",
    "auto",
]

MediaType = Literal[
    "auto-select",
    "rj45",
    "sfp",
]


class DynamicDhcpDistance(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    dynamic_dhcp_distance: Union[Variable, Global[int], Default[int]] = Default[int](value=1)


class InterfaceDynamicIPv4Address(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    dynamic: DynamicDhcpDistance


class StaticIPv4AddressConfig(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    primary_ip_address: StaticIPv4Address = Field(
        serialization_alias="staticIpV4AddressPrimary", validation_alias="staticIpV4AddressPrimary"
    )
    secondary_ip_address: Optional[StaticIPv4Address] = Field(
        serialization_alias="staticIpV4AddressSecondary", validation_alias="staticIpV4AddressSecondary", default=None
    )


class InterfaceStaticIPv4Address(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    static: StaticIPv4AddressConfig


class DynamicIPv6Dhcp(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    dhcp_client: Global[dict] = Field(
        serialization_alias="dhcpClient", validation_alias="dhcpClient", default=Global[dict](value={})
    )
    secondary_ipv6_address: Optional[List[StaticIPv6Address]] = Field(
        serialization_alias="secondaryIpV6Address", validation_alias="secondaryIpV6Address"
    )


class InterfaceDynamicIPv6Address(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    dynamic: DynamicIPv6Dhcp


class Dhcpv6Helper(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    ip_address: Union[Global[str], Variable] = Field(serialization_alias="ipAddress", validation_alias="ipAddress")
    vpn: Optional[Union[Global[int], Variable, Default[None]]] = None


class StaticIPv6AddressConfig(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    primary_ip_address: StaticIPv6Address = Field(
        serialization_alias="staticIpV6AddressPrimary", validation_alias="staticIpV6AddressPrimary"
    )
    secondary_ip_address: Optional[List[StaticIPv6Address]] = Field(
        serialization_alias="staticIpV6AddressSecondary", validation_alias="staticIpV6AddressSecondary", default=None
    )
    dhcp_helper_v6: Optional[List[Dhcpv6Helper]] = Field(
        serialization_alias="dhcpHelperV6", validation_alias="dhcpHelperV6", default=None
    )


class InterfaceStaticIPv6Address(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    static: StaticIPv6AddressConfig


class NatPool(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    range_start: Union[Variable, Global[str], Default[None]] = Field(
        serialization_alias="rangeStart", validation_alias="rangeStart"
    )
    range_end: Union[Variable, Global[str], Default[None]] = Field(
        serialization_alias="rangeEnd", validation_alias="rangeEnd"
    )
    prefix_length: Union[Variable, Global[int], Default[None]] = Field(
        serialization_alias="prefixLength", validation_alias="prefixLength"
    )
    overload: Union[Variable, Global[bool], Default[bool]] = Default[bool](value=True)


class StaticNat(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    source_ip: Union[Global[str], Variable] = Field(serialization_alias="sourceIp", validation_alias="sourceIp")

    translate_ip: Union[Global[str], Variable] = Field(
        serialization_alias="translateIp", validation_alias="translateIp"
    )
    static_nat_direction: Union[Global[Direction], Default[Direction]] = Field(
        serialization_alias="staticNatDirection",
        validation_alias="staticNatDirection",
        default=Default[Direction](value="inside"),
    )
    source_vpn: Union[Global[int], Variable, Default[int]] = Field(
        serialization_alias="sourceVpn", validation_alias="sourceVpn", default=Default[int](value=0)
    )


class NatAttributesIPv4(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    nat_type: Union[Global[NatType], Variable] = Field(serialization_alias="natType", validation_alias="natType")
    nat_pool: Optional[NatPool] = Field(serialization_alias="natPool", validation_alias="natPool", default=None)
    nat_loopback: Optional[Union[Global[str], Variable, Default[None]]] = Field(
        serialization_alias="natLoopbakc", validation_alias="natLoopbakc", default=None
    )
    udp_timeout: Union[Global[int], Variable, Default[int]] = Field(
        serialization_alias="udpTimeout", validation_alias="udpTimeout", default=Default[int](value=1)
    )
    tcp_timeout: Union[Global[int], Variable, Default[int]] = Field(
        serialization_alias="tcpTimeout", validation_alias="tcpTimeout", default=Default[int](value=1)
    )
    new_static_nat: Optional[List[StaticNat]] = Field(
        serialization_alias="newStaticNat", validation_alias="newStaticNat", default=None
    )


class NatAttributesIPv6(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    nat64: Optional[Union[Global[bool], Default[bool]]] = Default[bool](value=False)


class AclQos(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    shaping_rate: Optional[Union[Global[int], Variable, Default[None]]] = Field(
        serialization_alias="shapingRate", validation_alias="shapingRate", default=None
    )
    ipv4_acl_egress: Optional[Global[UUID]] = Field(
        serialization_alias="ipv4AclEgress", validation_alias="ipv4AclEgress", default=None
    )
    ipv4_acl_ingress: Optional[Global[UUID]] = Field(
        serialization_alias="ipv4AclIngress", validation_alias="ipv4AclIngress", default=None
    )
    ipv6_acl_egress: Optional[Global[UUID]] = Field(
        serialization_alias="ipv6AclEgress", validation_alias="ipv6AclEgress", default=None
    )
    ipv6_acl_ingress: Optional[Global[UUID]] = Field(
        serialization_alias="ipv6AclIngress", validation_alias="ipv6AclIngress", default=None
    )


class VrrpIPv6(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    group_id: Union[Variable, Global[int]] = Field(serialization_alias="groupId", validation_alias="groupId")
    priority: Union[Variable, Global[int], Default[int]] = Default[int](value=100)
    timer: Union[Variable, Global[int], Default[int]] = Default[int](value=1000)
    track_omp: Union[Global[bool], Default[bool]] = Field(
        serialization_alias="trackOmp", validation_alias="trackOmp", default=Default[bool](value=False)
    )
    ipv6: List[VrrpIPv6Address]


class VrrpIPv4(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    group_id: Union[Variable, Global[int]] = Field(serialization_alias="groupId", validation_alias="groupId")
    priority: Union[Variable, Global[int], Default[int]] = Default[int](value=100)
    timer: Union[Variable, Global[int], Default[int]] = Default[int](value=1000)
    track_omp: Union[Global[bool], Default[bool]] = Field(
        serialization_alias="trackOmp", validation_alias="trackOmp", default=Default[bool](value=False)
    )
    ip_address: Union[Global[str], Variable] = Field(serialization_alias="ipAddress", validation_alias="ipAddress")
    ip_address_secondary: Optional[List[StaticIPv4Address]] = Field(
        serialization_alias="ipAddressSecondary", validation_alias="ipAddressSecondary"
    )
    tloc_pref_change: Union[Global[bool], Default[bool]] = Field(
        serialization_alias="tlocPrefChange", validation_alias="tlocPrefChange", default=Default[bool](value=False)
    )
    tloc_pref_change_value: Optional[Union[Global[int], Default[None]]] = Field(
        serialization_alias="tlocPrefChangeValue", validation_alias="tlocPrefChangeValue", default=None
    )
    tracking_object: Optional[List[VrrpTrackingObject]] = Field(
        serialization_alias="trackingObject", validation_alias="trackingObject", default=None
    )


class Trustsec(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    enable_sgt_propagation: Union[Global[bool], Default[bool]] = Field(
        serialization_alias="enableSGTPropagation",
        validation_alias="enableSGTPropagation",
        default=Default[bool](value=False),
    )
    propagate: Optional[Union[Global[bool], Default[bool]]] = Default[bool](value=True)
    security_group_tag: Optional[Union[Global[int], Variable, Default[None]]] = Field(
        serialization_alias="securityGroupTag", validation_alias="securityGroupTag", default=None
    )
    enable_enforced_propagation: Union[Global[bool], Default[None]] = Field(
        serialization_alias="enableEnforcedPropagation", validation_alias="enableEnforcedPropagation"
    )
    enforced_security_group_tag: Union[Global[int], Variable, Default[None]] = Field(
        serialization_alias="enforcedSecurityGroupTag", validation_alias="enforcedSecurityGroupTag"
    )


class AdvancedAttributes(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    duplex: Optional[Union[Global[DuplexMode], Variable, Default[None]]] = None
    mac_address: Optional[Union[Global[str], Variable, Default[None]]] = Field(
        serialization_alias="macAddress", validation_alias="macAddress", default=None
    )
    ip_mtu: Union[Global[int], Variable, Default[int]] = Field(
        serialization_alias="ipMtu", validation_alias="ipMtu", default=Default[int](value=1500)
    )
    interface_mtu: Optional[Union[Global[int], Variable, Default[int]]] = Field(
        serialization_alias="intrfMtu", validation_alias="intrfMtu", default=Default[int](value=1500)
    )
    tcp_mss: Optional[Union[Global[int], Variable, Default[int]]] = Field(
        serialization_alias="tcpMss", validation_alias="tcpMss", default=None
    )
    speed: Optional[Union[Global[str], Variable, Default[None]]] = None
    arp_timeout: Union[Global[int], Variable, Default[int]] = Field(
        serialization_alias="arpTimeout", validation_alias="arpTimeout", default=Default[int](value=1200)
    )
    autonegotiate: Optional[Union[Global[bool], Variable, Default[bool]]] = None
    media_type: Optional[Union[Global[MediaType], Variable, Default[None]]] = Field(
        serialization_alias="mediaType", validation_alias="mediaType", default=None
    )
    load_interval: Union[Global[int], Variable, Default[int]] = Field(
        serialization_alias="loadInterval", validation_alias="loadInterval", default=Default[int](value=30)
    )
    tracker: Optional[Union[Global[str], Variable, Default[None]]] = None
    icmp_redirect_disable: Optional[Union[Global[bool], Variable, Default[bool]]] = Field(
        serialization_alias="icmpRedirectDisable",
        validation_alias="icmpRedirectDisable",
        default=Default[bool](value=True),
    )
    xconnect: Optional[Union[Global[str], Variable, Default[None]]] = None
    ip_directed_broadcast: Union[Global[bool], Variable, Default[bool]] = Field(
        serialization_alias="ipDirectedBroadcast",
        validation_alias="ipDirectedBroadcast",
        default=Default[bool](value=False),
    )


class InterfaceEthernetData(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    shutdown: Union[Global[bool], Variable, Default[bool]] = Default[bool](value=True)
    interface_name: Union[Global[str], Variable] = Field(
        serialization_alias="interfaceName", validation_alias="interfaceName"
    )
    description: Optional[Union[Global[str], Variable, Default[None]]] = None
    interface_ip_address: Union[InterfaceDynamicIPv4Address, InterfaceStaticIPv4Address] = Field(
        serialization_alias="intfIpAddress", validation_alias="intfIpAddress"
    )
    dhcp_helper: Optional[Union[Variable, Global[List[str]], Default[None]]] = Field(
        serialization_alias="dhcpHelper", validation_alias="dhcpHelper", default=None
    )
    interface_ipv6_address: Optional[Union[InterfaceDynamicIPv6Address, InterfaceStaticIPv6Address]] = Field(
        serialization_alias="intfIpV6Address", validation_alias="intfIpV6Address", default=None
    )
    nat: Union[Global[bool], Default[bool]] = Default[bool](value=False)
    nat_attributes_ipv4: Optional[NatAttributesIPv4] = Field(
        serialization_alias="natAttributesIpv4", validation_alias="natAttributesIpv4", default=None
    )
    nat_ipv6: Optional[Union[Global[bool], Default[bool]]] = Field(
        serialization_alias="natIpv6", validation_alias="natIpv6", default=Default[bool](value=False)
    )
    nat_attributes_ipv6: Optional[NatAttributesIPv6] = Field(
        serialization_alias="natAttributesIpv6", validation_alias="natAttributesIpv6", default=None
    )
    acl_qos: Optional[AclQos] = Field(serialization_alias="aclQos", validation_alias="aclQos", default=None)
    vrrp_ipv6: Optional[List[VrrpIPv6]] = Field(
        serialization_alias="vrrpIpv6", validation_alias="vrrpIpv6", default=None
    )
    vrrp: Optional[List[VrrpIPv4]] = None
    arp: Optional[List[Arp]] = None
    trustsec: Optional[Trustsec] = None
    advanced: AdvancedAttributes = AdvancedAttributes()


class InterfaceEthernetCreationPayload(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    name: str
    description: Optional[str] = None
    data: InterfaceEthernetData
    metadata: Optional[dict] = None



# Copyright 2023 Cisco Systems, Inc. and its affiliates

from enum import Enum
from pathlib import PurePath

from attr import define  # type: ignore
from clint.textui.progress import Bar as ProgressBar  # type: ignore
from requests_toolbelt.multipart.encoder import MultipartEncoder, MultipartEncoderMonitor  # type: ignore

from catalystwan.endpoints import CustomPayloadType, PreparedPayload
from catalystwan.endpoints.configuration_device_inventory import DeviceDetailsResponse
from catalystwan.exceptions import MultiplePersonalityError
from catalystwan.typed_list import DataSequence
from catalystwan.utils.personality import Personality


class Family(Enum):
    VEDGE = "vedge"
    VMANAGE = "vmanage"
    VEDGE_X86 = "vedge-x86"
    C8000V = "c8000v"


class VersionType(Enum):
    VMANAGE = "vmanage"
    REMOTE = "remote"


class DeviceType(Enum):
    CONTROLLER = "controller"
    VEDGE = "vedge"
    VMANAGE = "vmanage"


@define
class InstallSpecification:
    family: Family
    version_type: VersionType
    device_type: DeviceType


class InstallSpecHelper(Enum):
    """
    Container class for storage payload data for all personalities.
    It's created, because personality is not clearly connected to Family, VersionType
    or DeviceType
    """

    VMANAGE = InstallSpecification(Family.VMANAGE, VersionType.VMANAGE, DeviceType.VMANAGE)  # type: ignore
    VSMART = InstallSpecification(Family.VEDGE, VersionType.VMANAGE, DeviceType.CONTROLLER)  # type: ignore
    VBOND = InstallSpecification(Family.VEDGE, VersionType.VMANAGE, DeviceType.CONTROLLER)  # type: ignore
    VEDGE = InstallSpecification(Family.VEDGE, VersionType.VMANAGE, DeviceType.VEDGE)  # type: ignore
    REMOTE_VMANAGE = InstallSpecification(Family.VMANAGE, VersionType.REMOTE, DeviceType.VMANAGE)  # type: ignore
    REMOTE_VSMART = InstallSpecification(Family.VEDGE_X86, VersionType.REMOTE, DeviceType.CONTROLLER)  # type: ignore
    REMOTE_VBOND = InstallSpecification(Family.VEDGE_X86, VersionType.REMOTE, DeviceType.CONTROLLER)  # type: ignore
    REMOTE_VEDGE = InstallSpecification(Family.C8000V, VersionType.REMOTE, DeviceType.VEDGE)  # type: ignore


def get_install_specification(device: DeviceDetailsResponse, remote: bool = False):
    specification_container = {
        Personality.VMANAGE: InstallSpecHelper.REMOTE_VMANAGE.value if remote else InstallSpecHelper.VMANAGE.value,
        Personality.VBOND: InstallSpecHelper.REMOTE_VBOND.value if remote else InstallSpecHelper.VBOND.value,
        Personality.VSMART: InstallSpecHelper.REMOTE_VSMART.value if remote else InstallSpecHelper.VSMART.value,
        Personality.EDGE: InstallSpecHelper.REMOTE_VEDGE.value if remote else InstallSpecHelper.VEDGE.value,
    }
    return specification_container[device.personality]  # type: ignore


def validate_personality_homogeneity(devices: DataSequence[DeviceDetailsResponse]):
    personalities = set([device.personality for device in devices])
    if not len(personalities) == 1:
        raise MultiplePersonalityError(
            f"devices has got more than 1 personality, devices personalities: {personalities}"
        )


class SoftwarePackageUploadPayload(CustomPayloadType):
    def __init__(self, image_path: str):
        encoder = MultipartEncoder(
            fields={"file": (PurePath(image_path).name, open(image_path, "rb"), "application/x-gzip")}
        )
        callback = self._create_callback(encoder)
        monitor = MultipartEncoderMonitor(encoder, callback)
        self.payload = PreparedPayload(data=monitor, headers={"content-type": monitor.content_type})

    def _create_callback(self, encoder: MultipartEncoder):
        bar = ProgressBar(expected_size=encoder._calculate_length(), filled_char="=")

        def callback(monitor: MultipartEncoderMonitor):
            bar.show(monitor.bytes_read)

        return callback

    def prepared(self) -> PreparedPayload:
        return self.payload



# Copyright 2024 Cisco Systems, Inc. and its affiliates

from typing import Literal, Optional, Union
from uuid import UUID

from pydantic import BaseModel, ConfigDict, Field

from catalystwan.api.configuration_groups.parcel import Default, Global, Variable

IkeMode = Literal[
    "main",
    "aggresive",
]

IkeCiphersuite = Literal[
    "aes256-cbc-sha1",
    "aes256-cbc-sha2",
    "aes128-cbc-sha1",
    "aes128-cbc-sha2",
]

IkeGroup = Literal[
    "2",
    "14",
    "15",
    "16",
    "19",
    "20",
    "21",
    "24",
]

IpsecCiphersuite = Literal[
    "aes256-cbc-sha1",
    "aes256-cbc-sha384",
    "aes256-cbc-sha256",
    "aes256-cbc-sha512",
    "aes256-gcm",
    "null-sha1",
    "null-sha384",
    "null-sha256",
    "null-sha512",
]

PfsGroup = Literal[
    "group-1",
    "group-2",
    "group-5",
    "group-14",
    "group-15",
    "group-16",
    "group-19",
    "group-20",
    "group-21",
    "group-24",
    "none",
]

TunnelApplication = Literal[
    "none",
    "sig",
]

VrrpTrackerAction = Literal[
    "Decrement",
    "Shutdown",
]


class Arp(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    ip_address: Union[Variable, Global[str], Default[None]]
    mac_address: Union[Global[str], Variable]


class VrrpTrackingObject(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    tracker_id: Union[Default[None], Global[UUID]] = Field(
        serialization_alias="trackerId", validation_alias="trackerId"
    )
    tracker_action: Union[Global[VrrpTrackerAction], Variable] = Field(
        serialization_alias="trackerAction", validation_alias="trackerAction"
    )
    decrement_value: Optional[Union[Variable, Global[int]]] = Field(
        serialization_alias="decrementValue", validation_alias="decrementValue", default=None
    )


class VrrpIPv6Address(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    ipv6_link_local: Union[Global[str], Variable] = Field(
        serialization_alias="ipv6LinkLocal", validation_alias="ipv6LinkLocal"
    )
    prefix: Optional[Union[Global[str], Variable, Default[None]]] = None


class StaticIPv4Address(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    ip_address: Union[Variable, Global[str], Default[None]] = Field(
        serialization_alias="ipAddress", validation_alias="ipAddress"
    )
    subnet_mask: Union[Variable, Global[str], Default[None]] = Field(
        serialization_alias="subnetMask", validation_alias="subnetMask"
    )


class StaticIPv6Address(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    address: Union[Global[str], Variable]



# Copyright 2023 Cisco Systems, Inc. and its affiliates

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Protocol, Type, Union, overload
from uuid import UUID

from catalystwan.typed_list import DataSequence

if TYPE_CHECKING:
    from catalystwan.session import ManagerSession

from catalystwan.api.parcel_api import SDRoutingFullConfigParcelAPI
from catalystwan.endpoints.configuration.feature_profile.sdwan.policy_object import PolicyObjectFeatureProfile
from catalystwan.endpoints.configuration_feature_profile import SDRoutingConfigurationFeatureProfile
from catalystwan.models.configuration.feature_profile.common import (
    FeatureProfileCreationPayload,
    FeatureProfileCreationResponse,
    Parcel,
    ParcelCreationResponse,
)
from catalystwan.models.configuration.feature_profile.sdwan.policy_object import (
    POLICY_OBJECT_PAYLOAD_ENDPOINT_MAPPING,
    AnyPolicyObjectParcel,
    ApplicationListParcel,
    AppProbeParcel,
    ColorParcel,
    DataPrefixParcel,
    ExpandedCommunityParcel,
    FowardingClassParcel,
    FQDNDomainParcel,
    GeoLocationListParcel,
    IPSSignatureParcel,
    IPv6DataPrefixParcel,
    IPv6PrefixListParcel,
    LocalDomainParcel,
    PolicierParcel,
    PreferredColorGroupParcel,
    PrefixListParcel,
    ProtocolListParcel,
    SecurityApplicationListParcel,
    SecurityDataPrefixParcel,
    SecurityPortParcel,
    SecurityZoneListParcel,
    StandardCommunityParcel,
    TlocParcel,
    URLAllowParcel,
    URLBlockParcel,
)


class SDRoutingFeatureProfilesAPI:
    def __init__(self, session: ManagerSession):
        self.cli = SDRoutingCLIFeatureProfileAPI(session=session)
        self.policy_object = PolicyObjectFeatureProfileAPI(session=session)


class FeatureProfileAPI(Protocol):
    def init_parcels(self, fp_id: str) -> None:
        """
        Initialized parcel(s) associated with this feature profile
        """
        ...

    def create(self, name: str, description: str) -> FeatureProfileCreationResponse:
        """
        Creates feature profile
        """
        ...

    def delete(self, fp_id: str) -> None:
        """
        Deletes feature profile
        """
        ...


class SDRoutingCLIFeatureProfileAPI(FeatureProfileAPI):
    """
    SD-Routing CLI feature-profile APIs
    """

    def __init__(self, session: ManagerSession):
        self.session = session
        self.endpoint = SDRoutingConfigurationFeatureProfile(session)

    def init_parcels(self, fp_id: str) -> None:
        """
        Initialize CLI full-config parcel associated with this feature profile
        """
        self.full_config_parcel = SDRoutingFullConfigParcelAPI(session=self.session, fp_id=fp_id)

    def create(self, name: str, description: str) -> FeatureProfileCreationResponse:
        """
        Creates CLI feature profile
        """
        payload = FeatureProfileCreationPayload(name=name, description=description)

        return self.endpoint.create_cli_feature_profile(payload=payload)

    def delete(self, fp_id: str) -> None:
        """
        Deletes CLI feature-profile
        """
        self.endpoint.delete_cli_feature_profile(cli_fp_id=fp_id)


class PolicyObjectFeatureProfileAPI:
    """
    SDWAN Feature Profile Policy Object APIs
    """

    def __init__(self, session: ManagerSession):
        self.session = session
        self.endpoint = PolicyObjectFeatureProfile(session)

    @overload
    def get(self, profile_id: UUID, parcel_type: Type[ApplicationListParcel]) -> DataSequence[Parcel[Any]]:
        ...

    @overload
    def get(self, profile_id: UUID, parcel_type: Type[AppProbeParcel]) -> DataSequence[Parcel[Any]]:
        ...

    @overload
    def get(self, profile_id: UUID, parcel_type: Type[ColorParcel]) -> DataSequence[Parcel[Any]]:
        ...

    @overload
    def get(self, profile_id: UUID, parcel_type: Type[DataPrefixParcel]) -> DataSequence[Parcel[Any]]:
        ...

    @overload
    def get(self, profile_id: UUID, parcel_type: Type[ExpandedCommunityParcel]) -> DataSequence[Parcel[Any]]:
        ...

    @overload
    def get(self, profile_id: UUID, parcel_type: Type[FowardingClassParcel]) -> DataSequence[Parcel[Any]]:
        ...

    @overload
    def get(self, profile_id: UUID, parcel_type: Type[FQDNDomainParcel]) -> DataSequence[Parcel[Any]]:
        ...

    @overload
    def get(self, profile_id: UUID, parcel_type: Type[GeoLocationListParcel]) -> DataSequence[Parcel[Any]]:
        ...

    @overload
    def get(self, profile_id: UUID, parcel_type: Type[IPSSignatureParcel]) -> DataSequence[Parcel[Any]]:
        ...

    @overload
    def get(self, profile_id: UUID, parcel_type: Type[IPv6DataPrefixParcel]) -> DataSequence[Parcel[Any]]:
        ...

    @overload
    def get(self, profile_id: UUID, parcel_type: Type[IPv6PrefixListParcel]) -> DataSequence[Parcel[Any]]:
        ...

    @overload
    def get(self, profile_id: UUID, parcel_type: Type[LocalDomainParcel]) -> DataSequence[Parcel[Any]]:
        ...

    @overload
    def get(self, profile_id: UUID, parcel_type: Type[PolicierParcel]) -> DataSequence[Parcel[Any]]:
        ...

    @overload
    def get(self, profile_id: UUID, parcel_type: Type[PreferredColorGroupParcel]) -> DataSequence[Parcel[Any]]:
        ...

    @overload
    def get(self, profile_id: UUID, parcel_type: Type[PrefixListParcel]) -> DataSequence[Parcel[Any]]:
        ...

    @overload
    def get(self, profile_id: UUID, parcel_type: Type[ProtocolListParcel]) -> DataSequence[Parcel[Any]]:
        ...

    @overload
    def get(self, profile_id: UUID, parcel_type: Type[SecurityApplicationListParcel]) -> DataSequence[Parcel[Any]]:
        ...

    @overload
    def get(self, profile_id: UUID, parcel_type: Type[SecurityDataPrefixParcel]) -> DataSequence[Parcel[Any]]:
        ...

    @overload
    def get(self, profile_id: UUID, parcel_type: Type[SecurityPortParcel]) -> DataSequence[Parcel[Any]]:
        ...

    @overload
    def get(self, profile_id: UUID, parcel_type: Type[SecurityZoneListParcel]) -> DataSequence[Parcel[Any]]:
        ...

    @overload
    def get(self, profile_id: UUID, parcel_type: Type[StandardCommunityParcel]) -> DataSequence[Parcel[Any]]:
        ...

    @overload
    def get(self, profile_id: UUID, parcel_type: Type[TlocParcel]) -> DataSequence[Parcel[Any]]:
        ...

    @overload
    def get(self, profile_id: UUID, parcel_type: Type[URLAllowParcel]) -> DataSequence[Parcel[Any]]:
        ...

    @overload
    def get(self, profile_id: UUID, parcel_type: Type[URLBlockParcel]) -> DataSequence[Parcel[Any]]:
        ...

    # get by id

    @overload
    def get(
        self, profile_id: UUID, parcel_type: Type[ApplicationListParcel], parcel_id: UUID
    ) -> DataSequence[Parcel[Any]]:
        ...

    @overload
    def get(self, profile_id: UUID, parcel_type: Type[AppProbeParcel], parcel_id: UUID) -> DataSequence[Parcel[Any]]:
        ...

    @overload
    def get(self, profile_id: UUID, parcel_type: Type[ColorParcel], parcel_id: UUID) -> DataSequence[Parcel[Any]]:
        ...

    @overload
    def get(self, profile_id: UUID, parcel_type: Type[DataPrefixParcel], parcel_id: UUID) -> DataSequence[Parcel[Any]]:
        ...

    @overload
    def get(
        self, profile_id: UUID, parcel_type: Type[ExpandedCommunityParcel], parcel_id: UUID
    ) -> DataSequence[Parcel[Any]]:
        ...

    @overload
    def get(
        self, profile_id: UUID, parcel_type: Type[FowardingClassParcel], parcel_id: UUID
    ) -> DataSequence[Parcel[Any]]:
        ...

    @overload
    def get(self, profile_id: UUID, parcel_type: Type[FQDNDomainParcel], parcel_id: UUID) -> DataSequence[Parcel[Any]]:
        ...

    @overload
    def get(
        self, profile_id: UUID, parcel_type: Type[GeoLocationListParcel], parcel_id: UUID
    ) -> DataSequence[Parcel[Any]]:
        ...

    @overload
    def get(
        self, profile_id: UUID, parcel_type: Type[IPSSignatureParcel], parcel_id: UUID
    ) -> DataSequence[Parcel[Any]]:
        ...

    @overload
    def get(
        self, profile_id: UUID, parcel_type: Type[IPv6DataPrefixParcel], parcel_id: UUID
    ) -> DataSequence[Parcel[Any]]:
        ...

    @overload
    def get(
        self, profile_id: UUID, parcel_type: Type[IPv6PrefixListParcel], parcel_id: UUID
    ) -> DataSequence[Parcel[Any]]:
        ...

    @overload
    def get(self, profile_id: UUID, parcel_type: Type[LocalDomainParcel], parcel_id: UUID) -> DataSequence[Parcel[Any]]:
        ...

    @overload
    def get(self, profile_id: UUID, parcel_type: Type[PolicierParcel], parcel_id: UUID) -> DataSequence[Parcel[Any]]:
        ...

    @overload
    def get(
        self, profile_id: UUID, parcel_type: Type[PreferredColorGroupParcel], parcel_id: UUID
    ) -> DataSequence[Parcel[Any]]:
        ...

    @overload
    def get(self, profile_id: UUID, parcel_type: Type[PrefixListParcel], parcel_id: UUID) -> DataSequence[Parcel[Any]]:
        ...

    @overload
    def get(
        self, profile_id: UUID, parcel_type: Type[ProtocolListParcel], parcel_id: UUID
    ) -> DataSequence[Parcel[Any]]:
        ...

    @overload
    def get(
        self, profile_id: UUID, parcel_type: Type[SecurityApplicationListParcel], parcel_id: UUID
    ) -> DataSequence[Parcel[Any]]:
        ...

    @overload
    def get(
        self, profile_id: UUID, parcel_type: Type[SecurityDataPrefixParcel], parcel_id: UUID
    ) -> DataSequence[Parcel[Any]]:
        ...

    @overload
    def get(
        self, profile_id: UUID, parcel_type: Type[SecurityPortParcel], parcel_id: UUID
    ) -> DataSequence[Parcel[Any]]:
        ...

    @overload
    def get(
        self, profile_id: UUID, parcel_type: Type[SecurityZoneListParcel], parcel_id: UUID
    ) -> DataSequence[Parcel[Any]]:
        ...

    @overload
    def get(
        self, profile_id: UUID, parcel_type: Type[StandardCommunityParcel], parcel_id: UUID
    ) -> DataSequence[Parcel[Any]]:
        ...

    @overload
    def get(self, profile_id: UUID, parcel_type: Type[TlocParcel], parcel_id: UUID) -> DataSequence[Parcel[Any]]:
        ...

    @overload
    def get(self, profile_id: UUID, parcel_type: Type[URLAllowParcel], parcel_id: UUID) -> DataSequence[Parcel[Any]]:
        ...

    @overload
    def get(self, profile_id: UUID, parcel_type: Type[URLBlockParcel], parcel_id: UUID) -> DataSequence[Parcel[Any]]:
        ...

    def get(
        self,
        profile_id: UUID,
        parcel_type: Type[AnyPolicyObjectParcel],
        parcel_id: Union[UUID, None] = None,
    ) -> DataSequence[Parcel[Any]]:
        """
        Get all Policy Objects for selected profile_id and selected type or get one Policy Object given parcel id
        """

        policy_object_list_type = POLICY_OBJECT_PAYLOAD_ENDPOINT_MAPPING[parcel_type]
        if not parcel_id:
            return self.endpoint.get_all(profile_id=profile_id, policy_object_list_type=policy_object_list_type)
        parcel = self.endpoint.get_by_id(
            profile_id=profile_id, policy_object_list_type=policy_object_list_type, list_object_id=parcel_id
        )
        return DataSequence(Parcel, [parcel])

    def create(self, profile_id: UUID, payload: AnyPolicyObjectParcel) -> ParcelCreationResponse:
        """
        Create Policy Object for selected profile_id based on payload type
        """

        policy_object_list_type = POLICY_OBJECT_PAYLOAD_ENDPOINT_MAPPING[type(payload)]
        return self.endpoint.create(
            profile_id=profile_id, policy_object_list_type=policy_object_list_type, payload=payload
        )

    def update(self, profile_id: UUID, payload: AnyPolicyObjectParcel, list_object_id: UUID):
        """
        Update Policy Object for selected profile_id based on payload type
        """

        policy_type = POLICY_OBJECT_PAYLOAD_ENDPOINT_MAPPING[type(payload)]
        return self.endpoint.update(
            profile_id=profile_id, policy_object_list_type=policy_type, list_object_id=list_object_id, payload=payload
        )

    @overload
    def delete(self, profile_id: UUID, parcel_type: Type[ApplicationListParcel], list_object_id: UUID) -> None:
        ...

    @overload
    def delete(self, profile_id: UUID, parcel_type: Type[AppProbeParcel], list_object_id: UUID) -> None:
        ...

    @overload
    def delete(self, profile_id: UUID, parcel_type: Type[ColorParcel], list_object_id: UUID) -> None:
        ...

    @overload
    def delete(self, profile_id: UUID, parcel_type: Type[DataPrefixParcel], list_object_id: UUID) -> None:
        ...

    @overload
    def delete(self, profile_id: UUID, parcel_type: Type[ExpandedCommunityParcel], list_object_id: UUID) -> None:
        ...

    @overload
    def delete(self, profile_id: UUID, parcel_type: Type[FowardingClassParcel], list_object_id: UUID) -> None:
        ...

    @overload
    def delete(self, profile_id: UUID, parcel_type: Type[FQDNDomainParcel], list_object_id: UUID) -> None:
        ...

    @overload
    def delete(self, profile_id: UUID, parcel_type: Type[GeoLocationListParcel], list_object_id: UUID) -> None:
        ...

    @overload
    def delete(self, profile_id: UUID, parcel_type: Type[IPSSignatureParcel], list_object_id: UUID) -> None:
        ...

    @overload
    def delete(self, profile_id: UUID, parcel_type: Type[IPv6DataPrefixParcel], list_object_id: UUID) -> None:
        ...

    @overload
    def delete(self, profile_id: UUID, parcel_type: Type[IPv6PrefixListParcel], list_object_id: UUID) -> None:
        ...

    @overload
    def delete(self, profile_id: UUID, parcel_type: Type[LocalDomainParcel], list_object_id: UUID) -> None:
        ...

    @overload
    def delete(self, profile_id: UUID, parcel_type: Type[PolicierParcel], list_object_id: UUID) -> None:
        ...

    @overload
    def delete(self, profile_id: UUID, parcel_type: Type[PreferredColorGroupParcel], list_object_id: UUID) -> None:
        ...

    @overload
    def delete(self, profile_id: UUID, parcel_type: Type[PrefixListParcel], list_object_id: UUID) -> None:
        ...

    @overload
    def delete(self, profile_id: UUID, parcel_type: Type[ProtocolListParcel], list_object_id: UUID) -> None:
        ...

    @overload
    def delete(self, profile_id: UUID, parcel_type: Type[SecurityApplicationListParcel], list_object_id: UUID) -> None:
        ...

    @overload
    def delete(self, profile_id: UUID, parcel_type: Type[SecurityDataPrefixParcel], list_object_id: UUID) -> None:
        ...

    @overload
    def delete(self, profile_id: UUID, parcel_type: Type[SecurityPortParcel], list_object_id: UUID) -> None:
        ...

    @overload
    def delete(self, profile_id: UUID, parcel_type: Type[SecurityZoneListParcel], list_object_id: UUID) -> None:
        ...

    @overload
    def delete(self, profile_id: UUID, parcel_type: Type[StandardCommunityParcel], list_object_id: UUID) -> None:
        ...

    @overload
    def delete(self, profile_id: UUID, parcel_type: Type[TlocParcel], list_object_id: UUID) -> None:
        ...

    @overload
    def delete(self, profile_id: UUID, parcel_type: Type[URLAllowParcel], list_object_id: UUID) -> None:
        ...

    @overload
    def delete(self, profile_id: UUID, parcel_type: Type[URLBlockParcel], list_object_id: UUID) -> None:
        ...

    def delete(self, profile_id: UUID, parcel_type: Type[AnyPolicyObjectParcel], list_object_id: UUID) -> None:
        """
        Delete Policy Object for selected profile_id based on payload type
        """

        policy_object_list_type = POLICY_OBJECT_PAYLOAD_ENDPOINT_MAPPING[parcel_type]
        return self.endpoint.delete(
            profile_id=profile_id, policy_object_list_type=policy_object_list_type, list_object_id=list_object_id
        )



# Copyright 2023 Cisco Systems, Inc. and its affiliates

import unittest

from packaging.version import Version  # type: ignore
from parameterized import parameterized  # type: ignore

from catalystwan.version import NullVersion, parse_api_version, parse_vmanage_version


class TestNullVersion(unittest.TestCase):
    def test_null_version_properties(self):
        # Arrange, Act
        version = NullVersion()

        # Assert
        self.assertEqual(version.micro, 0)
        self.assertEqual(version.minor, 0)
        self.assertEqual(version.major, 0)
        self.assertEqual(version.base_version, "0")
        self.assertEqual(version.public, "NullVersion")
        self.assertEqual(version.local, None)
        self.assertEqual(version.dev, None)
        self.assertEqual(version.post, None)
        self.assertEqual(version.pre, None)
        self.assertEqual(version.release, (0,))
        self.assertEqual(version.epoch, 0)

        self.assertFalse(version.is_devrelease)
        self.assertFalse(version.is_postrelease)
        self.assertFalse(version.is_prerelease)

    def test_str(self):
        # Arrange, Act
        version = NullVersion()

        # Assert
        self.assertEqual(version.__str__(), "NullVersion")

    @parameterized.expand(
        [
            (Version("0"),),
            (Version("1.0.0"),),
            (Version("1.0.0-111"),),
            (Version("20.12"),),
        ]
    )
    def test_compare(self, version: Version):
        # Arrange, Act
        null_version = NullVersion()

        # Assert
        self.assertTrue(version > null_version)


class TestVersion(unittest.TestCase):
    @parameterized.expand(
        [
            ("1.0.0", Version("1.0.0")),
            ("0", Version("0")),
            ("20.12.0-914-xy", Version("20.12")),
            ("random_string", NullVersion()),
        ]
    )
    def test_parse_api_version(self, version: str, result: Version):
        # Arrange, Act
        parsed_version = parse_api_version(version)

        # Assert
        self.assertEqual(parsed_version, result)

    @parameterized.expand(
        [
            ("1.0.0", Version("1.0.0")),
            ("0", Version("0")),
            ("20.12.0-914-xy", Version("20.12.0-914")),
            ("abc-20.1.9", Version("20.1.9")),
            ("smart-li-20.13.999-3077", Version("20.13.999-3077")),
            ("random_string", NullVersion()),
        ]
    )
    def test_parse_vmanage_version(self, version: str, result: Version):
        # Arrange, Act
        parsed_version = parse_vmanage_version(version)

        # Assert
        self.assertEqual(parsed_version, result)

    def test_parse_api_version_object(self):
        # Arrange, Act
        parsed_version = parse_api_version("Not a Version!")

        # Assert
        self.assertTrue(isinstance(parsed_version, NullVersion))
        self.assertFalse(isinstance(Version("20.11"), NullVersion))


if __name__ == "__main__":
    unittest.main()



# Copyright 2023 Cisco Systems, Inc. and its affiliates

from enum import Enum
from pathlib import Path
from typing import ClassVar, List, Optional

from pydantic import ConfigDict, Field, field_validator

from catalystwan.api.templates.feature_template import FeatureTemplate, FeatureTemplateValidator


class Role(str, Enum):
    PRIMARY = "primary"
    SECONDARY = "secondary"


class Dns(FeatureTemplateValidator):
    dns_addr: Optional[str] = Field(default=None, json_schema_extra={"vmanage_key": "dns-addr"})
    role: Role = Role.PRIMARY
    model_config = ConfigDict(populate_by_name=True)


class DnsIpv6(FeatureTemplateValidator):
    dns_addr: Optional[str] = Field(default=None, json_schema_extra={"vmanage_key": "dns-addr"})
    role: Optional[Role] = Role.PRIMARY
    model_config = ConfigDict(populate_by_name=True)


class Host(FeatureTemplateValidator):
    hostname: str
    ip: List[str]


class SvcType(str, Enum):
    FW = "FW"
    IDS = "IDS"
    IDP = "IDP"
    NETSVC1 = "netsvc1"
    NETSVC2 = "netsvc2"
    NETSVC3 = "netsvc3"
    NETSVC4 = "netsvc4"
    TE = "TE"
    APPQOE = "appqoe"


class Service(FeatureTemplateValidator):
    svc_type: SvcType = Field(json_schema_extra={"vmanage_key": "svc-type"})
    address: Optional[List[str]] = None
    interface: Optional[str] = None
    track_enable: bool = Field(True, json_schema_extra={"vmanage_key": "track-enable"})
    model_config = ConfigDict(populate_by_name=True)

    @field_validator("track_enable")
    @classmethod
    def convert_to_string(cls, value):
        return str(value).lower()


class ServiceRouteService(str, Enum):
    SIG = "sig"


class ServiceRoute(FeatureTemplateValidator):
    prefix: str
    vpn: int
    service: ServiceRouteService = ServiceRouteService.SIG


class NextHop(FeatureTemplateValidator):
    address: Optional[str] = None
    distance: Optional[int] = 1


class NextHopWithTrack(FeatureTemplateValidator):
    address: Optional[str] = None
    distance: Optional[int] = 1
    tracker: str


class RouteInterface(FeatureTemplateValidator):
    interface_name: str = Field(json_schema_extra={"vmanage_key": "interface-name"})
    interface_next_hop: Optional[List[NextHop]] = Field(
        default=None, json_schema_extra={"vmanage_key": "interface-next-hop", "priority_order": ["address", "distance"]}
    )
    model_config = ConfigDict(populate_by_name=True)


class Routev4(FeatureTemplateValidator):
    prefix: Optional[str] = None
    next_hop: Optional[List[NextHop]] = Field(
        default=None, json_schema_extra={"vmanage_key": "next-hop", "priority_order": ["address", "distance"]}
    )
    next_hop_with_track: Optional[List[NextHopWithTrack]] = Field(
        default=None, json_schema_extra={"vmanage_key": "next-hop-with-track"}
    )
    route_interface: Optional[RouteInterface] = Field(
        default=None, json_schema_extra={"vmanage_key": "route-interface"}
    )
    null0: Optional[bool] = None
    distance: Optional[int] = None
    vpn: Optional[int] = None
    dhcp: Optional[bool] = None
    model_config = ConfigDict(populate_by_name=True)


class NextHopv6(FeatureTemplateValidator):
    address: str
    distance: Optional[int] = 1


class Nat(str, Enum):
    NAT64 = "NAT64"
    NAT66 = "NAT66"


class Routev6(FeatureTemplateValidator):
    prefix: str
    next_hop: Optional[List[NextHopv6]] = Field(default=None, json_schema_extra={"vmanage_key": "next-hop"})
    null0: Optional[bool] = None
    vpn: Optional[int] = None
    nat: Optional[Nat] = None
    model_config = ConfigDict(populate_by_name=True)


class GreRoute(FeatureTemplateValidator):
    prefix: str
    vpn: int
    interface: Optional[List[str]] = None


class IpsecRoute(FeatureTemplateValidator):
    prefix: str
    vpn: int
    interface: Optional[List[str]] = None


class AdvertiseProtocol(str, Enum):
    BGP = "bgp"
    OSPF = "ospf"
    OSPFV3 = "ospfv3"
    CONNECTED = "connected"
    STATIC = "static"
    NETWORK = "network"
    AGGREGATE = "aggregate"
    EIGRP = "eigrp"
    LISP = "lisp"
    ISIS = "isis"


class AdvertiseProtocolSubType(str, Enum):
    EXTERNAL = "external"


class Region(str, Enum):
    CORE = "core"
    ACCESS = "access"


class PrefixList(FeatureTemplateValidator):
    prefix_entry: str = Field(json_schema_extra={"vmanage_key": "prefix-entry"})
    aggregate_only: Optional[bool] = Field(default=None, json_schema_extra={"vmanage_key": "aggregate-only"})
    region: Optional[Region]
    model_config = ConfigDict(populate_by_name=True)


class Advertise(FeatureTemplateValidator):
    protocol: AdvertiseProtocol
    route_policy: Optional[str] = Field(default=None, json_schema_extra={"vmanage_key": "route-policy"})
    protocol_sub_type: Optional[List[AdvertiseProtocolSubType]] = Field(
        default=None, json_schema_extra={"vmanage_key": "protocol-sub-type"}
    )
    prefix_list: Optional[List[PrefixList]] = Field(default=None, json_schema_extra={"vmanage_key": "prefix-list"})
    model_config = ConfigDict(populate_by_name=True)


class Ipv6AdvertiseProtocol(str, Enum):
    BGP = "bgp"
    OSPF = "ospf"
    CONNECTED = "connected"
    STATIC = "static"
    NETWORK = "network"
    AGGREGATE = "aggregate"


class Ipv6AdvertiseProtocolSubType(str, Enum):
    EXTERNAL = "external"


class Ipv6Advertise(FeatureTemplateValidator):
    protocol: Ipv6AdvertiseProtocol
    route_policy: Optional[str] = Field(default=None, json_schema_extra={"vmanage_key": "route-policy"})
    protocol_sub_type: Optional[List[Ipv6AdvertiseProtocolSubType]] = Field(
        default=None, json_schema_extra={"vmanage_key": "protocol-sub-type"}
    )
    prefix_list: Optional[List[PrefixList]] = Field(default=None, json_schema_extra={"vmanage_key": "prefix-list"})
    model_config = ConfigDict(populate_by_name=True)


class LeakFromGlobalProtocol(str, Enum):
    ALL = "all"
    STATIC = "static"
    MOBILE = "mobile"
    CONNECTED = "connected"
    RIP = "rip"
    ODR = "odr"


class Pool(FeatureTemplateValidator):
    name: str
    start_address: str = Field(json_schema_extra={"vmanage_key": "start-address"})
    end_address: str = Field(json_schema_extra={"vmanage_key": "end-address"})
    overload: Optional[bool] = None
    leak_from_global: bool
    leak_from_global_protocol: LeakFromGlobalProtocol
    leak_to_global: bool
    model_config = ConfigDict(populate_by_name=True)


class Direction(str, Enum):
    INSIDE = "inside"
    OUTSIDE = "outside"


class Overload(str, Enum):
    TRUE = "true"
    FALSE = "false"


class Natpool(FeatureTemplateValidator):
    name: int
    prefix_length: Optional[int] = Field(default=None, json_schema_extra={"vmanage_key": "prefix-length"})
    range_start: str = Field(default=None, json_schema_extra={"vmanage_key": "range-start"})
    range_end: Optional[str] = Field(default=None, json_schema_extra={"vmanage_key": "range-end"})
    overload: Overload = Overload.TRUE
    direction: Direction
    tracker_id: Optional[int] = Field(default=None, json_schema_extra={"vmanage_key": "tracker-id"})
    model_config = ConfigDict(populate_by_name=True)


class StaticNatDirection(str, Enum):
    INSIDE = "inside"
    OUTSIDE = "outside"


class Static(FeatureTemplateValidator):
    pool_name: Optional[int] = Field(default=None, json_schema_extra={"vmanage_key": "pool-name"})
    source_ip: Optional[str] = Field(default=None, json_schema_extra={"vmanage_key": "source-ip"})
    translate_ip: Optional[str] = Field(default=None, json_schema_extra={"vmanage_key": "translate-ip"})
    static_nat_direction: StaticNatDirection = Field(json_schema_extra={"vmanage_key": "static-nat-direction"})
    tracker_id: Optional[int] = Field(default=None, json_schema_extra={"vmanage_key": "tracker-id"})
    model_config = ConfigDict(populate_by_name=True)


class SubnetStatic(FeatureTemplateValidator):
    source_ip_subnet: str = Field(json_schema_extra={"vmanage_key": "source-ip-subnet"})
    translate_ip_subnet: str = Field(json_schema_extra={"vmanage_key": "translate-ip-subnet"})
    prefix_length: int = Field(json_schema_extra={"vmanage_key": "prefix-length"})
    static_nat_direction: StaticNatDirection = Field(json_schema_extra={"vmanage_key": "static-nat-direction"})
    tracker_id: Optional[int] = Field(default=None, json_schema_extra={"vmanage_key": "tracker-id"})
    model_config = ConfigDict(populate_by_name=True)


class Proto(str, Enum):
    TCP = "tcp"
    UDP = "udp"


class PortForward(FeatureTemplateValidator):
    pool_name: Optional[int] = Field(default=None, json_schema_extra={"vmanage_key": "pool-name"})
    source_port: int = Field(json_schema_extra={"vmanage_key": "source-port"})
    translate_port: int = Field(json_schema_extra={"vmanage_key": "translate-port"})
    source_ip: str = Field(json_schema_extra={"vmanage_key": "source-ip"})
    translate_ip: str = Field(json_schema_extra={"vmanage_key": "translate-ip"})
    proto: Proto
    model_config = ConfigDict(populate_by_name=True)


class RouteImportProtocol(str, Enum):
    STATIC = "static"
    CONNECTED = "connected"
    BGP = "bgp"
    OSPF = "ospf"


class RouteImportProtocolSubType(str, Enum):
    EXTERNAL = "external"


class RouteImportRedistributeProtocol(str, Enum):
    BGP = "bgp"
    EIGRP = "eigrp"
    OSPF = "ospf"


class RouteImportRedistribute(FeatureTemplateValidator):
    protocol: RouteImportRedistributeProtocol
    route_policy: Optional[str] = Field(default=None, json_schema_extra={"vmanage_key": "route-policy"})
    model_config = ConfigDict(populate_by_name=True)


class RouteImport(FeatureTemplateValidator):
    protocol: RouteImportProtocol
    protocol_sub_type: List[RouteImportProtocolSubType] = Field(json_schema_extra={"vmanage_key": "protocol-sub-type"})
    route_policy: Optional[str] = Field(default=None, json_schema_extra={"vmanage_key": "route-policy"})
    redistribute: Optional[List[RouteImportRedistribute]] = None
    model_config = ConfigDict(populate_by_name=True)


class RouteImportFromProtocol(str, Enum):
    STATIC = "static"
    CONNECTED = "connected"
    BGP = "bgp"
    OSPF = "ospf"
    EIGRP = "eigrp"


class RouteImportFromProtocolSubType(str, Enum):
    EXTERNAL = "external"


class RouteImportFromRedistributeProtocol(str, Enum):
    BGP = "bgp"
    EIGRP = "eigrp"
    OSPF = "ospf"


class RouteImportFromRedistribute(FeatureTemplateValidator):
    protocol: RouteImportFromRedistributeProtocol
    route_policy: Optional[str] = Field(default=None, json_schema_extra={"vmanage_key": "route-policy"})
    model_config = ConfigDict(populate_by_name=True)


class RouteImportFrom(FeatureTemplateValidator):
    source_vpn: int = Field(json_schema_extra={"vmanage_key": "source-vpn"})
    protocol: RouteImportFromProtocol
    protocol_sub_type: List[RouteImportFromProtocolSubType] = Field(
        json_schema_extra={"vmanage_key": "protocol-sub-type"}
    )
    route_policy: Optional[str] = Field(default=None, json_schema_extra={"vmanage_key": "route-policy"})
    redistribute: Optional[List[RouteImportFromRedistribute]] = None
    model_config = ConfigDict(populate_by_name=True)


class RouteExportProtocol(str, Enum):
    STATIC = "static"
    CONNECTED = "connected"
    BGP = "bgp"
    EIGRP = "eigrp"
    OSPF = "ospf"


class RouteExportProtocolSubType(str, Enum):
    EXTERNAL = "external"


class RouteExportRedistributeProtocol(str, Enum):
    BGP = "bgp"
    OSPF = "ospf"


class RouteExportRedistribute(FeatureTemplateValidator):
    protocol: RouteExportRedistributeProtocol
    route_policy: Optional[str] = Field(default=None, json_schema_extra={"vmanage_key": "route-policy"})
    model_config = ConfigDict(populate_by_name=True)


class RouteExport(FeatureTemplateValidator):
    protocol: RouteExportProtocol
    protocol_sub_type: List[RouteExportProtocolSubType] = Field(json_schema_extra={"vmanage_key": "protocol-sub-type"})
    route_policy: Optional[str] = Field(default=None, json_schema_extra={"vmanage_key": "route-policy"})
    redistribute: Optional[List[RouteExportRedistribute]]
    model_config = ConfigDict(populate_by_name=True)


class CiscoVPNModel(FeatureTemplate):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    vpn_id: int = Field(default=0, json_schema_extra={"vmanage_key": "vpn-id"})
    vpn_name: Optional[str] = Field(default=None, json_schema_extra={"vmanage_key": "name"})
    tenant_vpn_id: Optional[int] = Field(default=None, json_schema_extra={"vmanage_key": "tenant-vpn-id"})
    org_name: Optional[str] = Field(default=None, json_schema_extra={"vmanage_key": "org-name"})
    omp_admin_distance_ipv4: Optional[int] = Field(
        default=None, json_schema_extra={"vmanage_key": "omp-admin-distance-ipv4"}
    )
    omp_admin_distance_ipv6: Optional[int] = Field(
        default=None, json_schema_extra={"vmanage_key": "omp-admin-distance-ipv6"}
    )
    dns: Optional[List[Dns]] = None
    dns_ipv6: Optional[List[DnsIpv6]] = Field(default=None, json_schema_extra={"vmanage_key": "dns-ipv6"})
    layer4: Optional[bool] = Field(default=None, json_schema_extra={"data_path": ["ecmp-hash-key"]})
    host: Optional[List[Host]] = Field(default=None, json_schema_extra={"priority_order": ["hostname", "ip"]})
    service: Optional[List[Service]] = None
    service_route: Optional[List[ServiceRoute]] = Field(
        default=None, json_schema_extra={"data_path": ["ip"], "vmanage_key": "service-route"}
    )
    route_v4: Optional[List[Routev4]] = Field(
        default=None,
        json_schema_extra={
            "data_path": ["ip"],
            "vmanage_key": "route",
            "priority_order": ["prefix", "next-hop", "next-hop-with-track"],
        },
    )
    route_v6: Optional[List[Routev6]] = Field(
        default=None, json_schema_extra={"data_path": ["ipv6"], "vmanage_key": "route"}
    )
    gre_route: Optional[List[GreRoute]] = Field(
        default=None, json_schema_extra={"data_path": ["ip"], "vmanage_key": "gre-route"}
    )
    ipsec_route: Optional[List[IpsecRoute]] = Field(
        default=None, json_schema_extra={"data_path": ["ip"], "vmanage_key": "ipsec-route"}
    )
    advertise: Optional[List[Advertise]] = Field(default=None, json_schema_extra={"data_path": ["omp"]})
    ipv6_advertise: Optional[List[Ipv6Advertise]] = Field(
        default=None, json_schema_extra={"data_path": ["omp"], "vmanage_key": "ipv6-advertise"}
    )
    pool: Optional[List[Pool]] = Field(default=None, json_schema_extra={"data_path": ["nat64", "v4"]})
    natpool: Optional[List[Natpool]] = Field(default=None, json_schema_extra={"data_path": ["nat"]})
    static: Optional[List[Static]] = Field(default=None, json_schema_extra={"data_path": ["nat"]})
    subnet_static: Optional[List[SubnetStatic]] = Field(
        default=None, json_schema_extra={"data_path": ["nat"], "vmanage_key": "subnet-static"}
    )
    port_forward: Optional[List[PortForward]] = Field(
        default=None, json_schema_extra={"data_path": ["nat"], "vmanage_key": "port-forward"}
    )
    route_import: Optional[List[RouteImport]] = Field(default=None, json_schema_extra={"vmanage_key": "route-import"})
    route_import_from: Optional[List[RouteImportFrom]] = Field(
        default=None, json_schema_extra={"vmanage_key": "route-import-from"}
    )
    route_export: Optional[List[RouteExport]] = Field(default=None, json_schema_extra={"vmanage_key": "route-export"})

    payload_path: ClassVar[Path] = Path(__file__).parent / "DEPRECATED"
    type: ClassVar[str] = "cisco_vpn"

    def generate_vpn_id(self, session):
        if self.vpn_id not in [0, 512]:
            payload = {"resourcePoolDataType": "vpn", "tenantId": self.org_name, "tenantVpn": self.vpn_id}
            url = "/dataservice/resourcepool/resource/vpn"
            response = session.put(url=url, json=payload).json()
            self.vpn_id = response["deviceVpn"]



# Copyright 2023 Cisco Systems, Inc. and its affiliates

# mypy: disable-error-code="empty-body"
from uuid import UUID

from catalystwan.endpoints import JSON, APIEndpoints, delete, get, post, put
from catalystwan.models.policy.security import (
    AnySecurityPolicy,
    SecurityPolicyEditResponse,
    SecurityPolicyInfoRoot,
    SecurityPolicyRoot,
)
from catalystwan.typed_list import DataSequence


class ConfigurationSecurityTemplatePolicy(APIEndpoints):
    @post("/template/policy/security")
    def create_security_template(self, payload: AnySecurityPolicy) -> None:
        ...

    @delete("/template/policy/security/{id}")
    def delete_security_template(self, id: UUID, payload: JSON = {}) -> None:
        ...

    @put("/template/policy/security/{id}")
    def edit_security_template(self, id: UUID, payload: AnySecurityPolicy) -> SecurityPolicyEditResponse:
        # PUT /template/policy/security/{policyId}
        ...

    def edit_template_with_lenient_lock(self):
        # PUT /template/policy/security/staging/{policyId}
        ...

    def generate_security_policy_summary(self):
        # GET /template/policy/security/summary
        ...

    @get("/template/policy/security", "data")
    def generate_security_template_list(self) -> DataSequence[SecurityPolicyInfoRoot]:
        ...

    def get_device_list_by_id(self):
        # GET /template/policy/security/devices/{policyId}
        ...

    def get_security_policy_device_list(self):
        # GET /template/policy/security/devices
        ...

    @get("/template/policy/security/definition/{id}")
    def get_security_template(self, id: UUID) -> SecurityPolicyRoot:
        ...

    def get_security_templates_for_device(self):
        # GET /template/policy/security/{deviceModel}
        ...



# Copyright 2024 Cisco Systems, Inc. and its affiliates

from typing import List, Literal, Optional, Union
from uuid import UUID

from pydantic import AliasPath, BaseModel, ConfigDict, Field, field_validator

from catalystwan.api.configuration_groups.parcel import Global, _ParcelBase, as_global

SLAClassCriteria = Literal[
    "loss",
    "latency",
    "jitter",
    "loss-latency",
    "loss-jitter",
    "latency-loss",
    "latency-jitter",
    "jitter-latency",
    "jitter-loss",
    "loss-latency-jitter",
    "loss-jitter-latency",
    "latency-loss-jitter",
    "latency-jitter-loss",
    "jitter-latency-loss",
    "jitter-loss-latency",
]


def check_latency_ms(cls, latency: Optional[Global]):
    if latency is not None:
        assert 1 <= latency.value <= 1000
        return latency


def check_loss_percent(cls, loss: Optional[Global]):
    if loss is not None:
        assert 0 <= loss.value <= 100
    return loss


def check_jitter_ms(cls, jitter: Optional[Global]):
    if jitter is not None:
        assert 1 <= jitter.value <= 1000
    return jitter


class SLAAppProbeClass(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    ref_id: Global[UUID] = Field(serialization_alias="refId", validation_alias="refId")


class FallbackBestTunnel(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    criteria: Global[SLAClassCriteria]
    jitter_variance: Optional[Global[int]] = Field(
        default=None,
        serialization_alias="jitterVariance",
        validation_alias="jitterVariance",
        description="jitter variance in ms",
    )
    latency_variance: Optional[Global[int]] = Field(
        default=None,
        serialization_alias="latencyVariance",
        validation_alias="latencyVariance",
        description="latency variance in ms",
    )
    loss_variance: Optional[Global[int]] = Field(
        default=None,
        serialization_alias="lossVariance",
        validation_alias="lossVariance",
        description="loss variance as percentage",
    )
    # validators
    _jitter_validator = field_validator("jitter_variance")(check_jitter_ms)
    _latency_validator = field_validator("latency_variance")(check_latency_ms)
    _loss_validator = field_validator("loss_variance")(check_loss_percent)

    def add_criteria(
        self, jitter_variance: Union[int, None], latency_variance: Union[int, None], loss_variance: Union[int, None]
    ):
        expected_criteria = []
        if jitter_variance:
            self.jitter_variance = Global(value=jitter_variance)
            expected_criteria.append("jitter")
        if latency_variance:
            self.latency_variance = Global(value=latency_variance)
            expected_criteria.append("latency")
        if loss_variance:
            self.loss_variance = Global(value=loss_variance)
            expected_criteria.append("loss")
        for e in expected_criteria:
            if e not in self.criteria.value:
                raise ValueError(f"Criteria {e} is not in configured criteria {self.criteria.value}")


class SLAClassListEntry(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    latency: Optional[Global[int]] = None
    loss: Optional[Global[int]] = None
    jitter: Optional[Global[int]] = None
    app_probe_class: Optional[SLAAppProbeClass] = Field(
        validation_alias="appProbeClass", serialization_alias="appProbeClass"
    )
    fallback_best_tunnel: Optional[FallbackBestTunnel] = Field(
        default=None, validation_alias="fallbackBestTunnel", serialization_alias="fallbackBestTunnel"
    )

    # validators
    _jitter_validator = field_validator("jitter")(check_jitter_ms)
    _latency_validator = field_validator("latency")(check_latency_ms)
    _loss_validator = field_validator("loss")(check_loss_percent)


class SLAClassParcel(_ParcelBase):
    entries: List[SLAClassListEntry] = Field(default=[], validation_alias=AliasPath("data", "entries"))

    def add_entry(
        self,
        app_probe_class_id: UUID,
        loss: Optional[int] = None,
        jitter: Optional[int] = None,
        latency: Optional[int] = None,
    ):
        self.entries.append(
            SLAClassListEntry(
                app_probe_class=SLAAppProbeClass(ref_id=as_global(app_probe_class_id)),
                loss=as_global(loss) if loss is not None else None,
                jitter=as_global(jitter) if jitter is not None else None,
                latency=as_global(latency) if latency is not None else None,
            )
        )

    def add_fallback(
        self,
        criteria: SLAClassCriteria,
        jitter_variance: Optional[int] = None,
        latency_variance: Optional[int] = None,
        loss_variance: Optional[int] = None,
    ):
        fallback = FallbackBestTunnel(
            criteria=as_global(criteria, SLAClassCriteria),
            jitter_variance=as_global(jitter_variance) if jitter_variance is not None else None,
            latency_variance=as_global(latency_variance) if latency_variance is not None else None,
            loss_variance=as_global(loss_variance) if loss_variance is not None else None,
        )
        self.entries[0].fallback_best_tunnel = fallback



# Copyright 2023 Cisco Systems, Inc. and its affiliates

from enum import Enum
from pathlib import Path
from typing import ClassVar, List, Optional

from pydantic import ConfigDict, Field

from catalystwan.api.templates.feature_template import FeatureTemplate, FeatureTemplateValidator


class User(FeatureTemplateValidator):
    name: str
    password: Optional[str] = None
    secret: Optional[str] = None
    privilege: Optional[str] = None
    pubkey_chain: List[str] = Field(default=[], json_schema_extra={"vmanage_key": "pubkey-chain", "vip_type": "ignore"})


class RadiusServer(FeatureTemplateValidator):
    model_config = ConfigDict(populate_by_name=True, coerce_numbers_to_str=True)

    address: str
    auth_port: int = Field(default=1812, json_schema_extra={"vmanage_key": "auth-port"})
    acct_port: int = Field(default=1813, json_schema_extra={"vmanage_key": "acct-port"})
    timeout: int = Field(default=5)
    retransmit: int = 3
    key: str
    secret_key: Optional[str] = Field(default=None, json_schema_extra={"vmanage_key": "secret-key"})
    key_enum: Optional[str] = Field(default=None, json_schema_extra={"vmanage_key": "key-enum"})
    key_type: Optional[str] = Field(default=None, json_schema_extra={"vmanage_key": "key-type"})


class RadiusGroup(FeatureTemplateValidator):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    group_name: str = Field(json_schema_extra={"vmanage_key": "group-name"})
    vpn: Optional[int] = None
    source_interface: Optional[str] = Field(json_schema_extra={"vmanage_key": "source-interface"})
    server: List[RadiusServer] = []


class DomainStripping(str, Enum):
    YES = "yes"
    NO = "no"
    RIGHT_TO_LEFT = "right-to-left"


class TacacsServer(FeatureTemplateValidator):
    model_config = ConfigDict(populate_by_name=True)

    address: str
    port: int = 49
    timeout: int = Field(default=5)
    key: str
    secret_key: Optional[str] = Field(default=None, json_schema_extra={"vmanage_key": "secret-key"})
    key_enum: Optional[str] = Field(default=None, json_schema_extra={"vmanage_key": "key-enum"})


class TacacsGroup(FeatureTemplateValidator):
    model_config = ConfigDict(populate_by_name=True)

    group_name: str = Field(json_schema_extra={"vmanage_key": "group-name"})
    vpn: int = 0
    source_interface: Optional[str] = Field(default=None, json_schema_extra={"vmanage_key": "source-interface"})
    server: List[TacacsServer] = []


class CiscoAAAModel(FeatureTemplate):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    user: Optional[List[User]] = None
    authentication_group: bool = Field(default=False, json_schema_extra={"vmanage_key": "authentication_group"})
    accounting_group: bool = True
    radius: Optional[List[RadiusGroup]] = None
    domain_stripping: Optional[DomainStripping] = Field(
        default=None, json_schema_extra={"vmanage_key": "domain-stripping"}
    )
    port: int = 1700
    tacacs: Optional[List[TacacsGroup]] = None
    server_auth_order: str = Field(default="local", json_schema_extra={"vmanage_key": "server-auth-order"})

    payload_path: ClassVar[Path] = Path(__file__).parent / "DEPRECATED"
    type: ClassVar[str] = "cedge_aaa"



# Copyright 2023 Cisco Systems, Inc. and its affiliates

# mypy: disable-error-code="empty-body"
from typing import Literal, Optional

from pydantic import BaseModel

from catalystwan.endpoints import APIEndpoints, get
from catalystwan.typed_list import DataSequence

TenancyModes = Literal["SingleTenant", "MultiTenant"]


class TenancyMode(BaseModel):
    mode: TenancyModes
    deploymentmode: str
    domain: Optional[str] = None
    clusterid: Optional[str] = None


class VManageDetails(BaseModel):
    service: str
    enabled: bool
    status: str


class ClusterManagement(APIEndpoints):
    def add_or_update_user_credentials(self):
        # POST /clusterManagement/userCreds
        ...

    def add_vmanage(self):
        # POST /clusterManagement/setup
        ...

    def check_if_cluster_locked(self):
        # GET /clusterManagement/clusterLocked
        ...

    def configure_vmanage(self):
        # POST /clusterManagement/configure
        ...

    def edit_vmanage(self):
        # PUT /clusterManagement/setup
        ...

    def get_cluster_workflow_version(self):
        # GET /clusterManagement/clusterworkflow/version
        ...

    def get_configured_ip_list(self):
        # GET /clusterManagement/iplist/{vmanageID}
        ...

    def get_connected_devices(self):
        # GET /clusterManagement/connectedDevices/{vmanageIP}
        ...

    def get_connected_devices_per_tenant(self):
        # GET /clusterManagement/{tenantId}/connectedDevices/{vmanageIP}
        ...

    @get("/clusterManagement/tenancy/mode", "data")
    def get_tenancy_mode(self) -> TenancyMode:
        ...

    def get_tenants_list(self):
        # GET /clusterManagement/tenantList
        ...

    @get("/clusterManagement/vManage/details/{vmanageIP}", "data")
    def get_vmanage_details(self, vmanageIP: str) -> DataSequence[VManageDetails]:
        ...

    def health_details(self):
        # GET /clusterManagement/health/details
        ...

    def health_status_info(self):
        # GET /clusterManagement/health/status
        ...

    def health_summary(self):
        # GET /clusterManagement/health/summary
        ...

    def is_cluster_ready(self):
        # GET /clusterManagement/isready
        ...

    def list_vmanages(self):
        # GET /clusterManagement/list
        ...

    def node_properties(self):
        # GET /clusterManagement/nodeProperties
        ...

    def perform_replication_and_rebalance_of_kafka_partitions(self):
        # PUT /clusterManagement/replicateAndRebalance
        ...

    def remove_vmanage(self):
        # POST /clusterManagement/remove
        ...

    def set_tenancy_mode(self):
        # POST /clusterManagement/tenancy/mode
        ...



# Copyright 2024 Cisco Systems, Inc. and its affiliates

# mypy: disable-error-code="empty-body"
from uuid import UUID

from catalystwan.endpoints import APIEndpoints, delete, get, post, put
from catalystwan.models.policy.lists import RegionList
from catalystwan.models.policy.policy_list import InfoTag, PolicyListId, PolicyListInfo, PolicyListPreview
from catalystwan.typed_list import DataSequence


class RegionListEditPayload(RegionList, PolicyListId):
    pass


class RegionListInfo(RegionList, PolicyListInfo):
    pass


class ConfigurationPolicyRegionList(APIEndpoints):
    @post("/template/policy/list/region")
    def create_policy_list(self, payload: RegionList) -> PolicyListId:
        ...

    @delete("/template/policy/list/region/{id}")
    def delete_policy_list(self, id: UUID) -> None:
        ...

    @delete("/template/policy/list/region")
    def delete_policy_lists_with_info_tag(self, params: InfoTag) -> None:
        ...

    @put("/template/policy/list/region/{id}")
    def edit_policy_list(self, id: UUID, payload: RegionListEditPayload) -> None:
        ...

    @get("/template/policy/list/region/{id}")
    def get_lists_by_id(self, id: UUID) -> RegionListInfo:
        ...

    @get("/template/policy/list/region", "data")
    def get_policy_lists(self) -> DataSequence[RegionListInfo]:
        ...

    @get("/template/policy/list/region/filtered", "data")
    def get_policy_lists_with_info_tag(self, params: InfoTag) -> DataSequence[RegionListInfo]:
        ...

    @post("/template/policy/list/region/preview")
    def preview_policy_list(self, payload: RegionList) -> PolicyListPreview:
        ...

    @get("/template/policy/list/region/preview/{id}")
    def preview_policy_list_by_id(self, id: UUID) -> PolicyListPreview:
        ...



# Copyright 2023 Cisco Systems, Inc. and its affiliates

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, ClassVar, List, Optional

from pydantic import BaseModel, ConfigDict

from catalystwan.api.templates.feature_template import FeatureTemplate
from catalystwan.endpoints.monitoring.device_details import Tier as TierInfo
from catalystwan.models.tenant import Tenant as TenantInfo

if TYPE_CHECKING:
    from catalystwan.session import ManagerSession


class Tenant(BaseModel):
    """Tenant definition without TLOCs."""

    organization_name: str
    tier_name: str
    tenant_info: Optional[TenantInfo] = None
    tier_info: Optional[TierInfo] = None


class TenantModel(FeatureTemplate):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    type: ClassVar[str] = "tenant"  # Tenant
    payload_path: ClassVar[Path] = Path(__file__).parent / "tenant.json.j2"

    tenants: List[Tenant] = []

    def generate_payload(self, session: ManagerSession) -> str:
        tenant_infos = session.endpoints.tenant_management.get_all_tenants()
        tier_infos = session.endpoints.monitoring_device_details.get_tiers()

        for tenant in self.tenants:
            tenant.tier_info = tier_infos.filter(name=tenant.tier_name).single_or_default()
            tenant.tenant_info = tenant_infos.filter(org_name=tenant.organization_name).single_or_default()

        return super().generate_payload(session)



# Copyright 2024 Cisco Systems, Inc. and its affiliates

from ipaddress import IPv4Address
from typing import Any, List, Literal, Optional, Union, overload
from uuid import UUID

from pydantic import ConfigDict, Field
from typing_extensions import Annotated

from catalystwan.models.common import TLOCColor
from catalystwan.models.policy.lists_entries import EncapType
from catalystwan.models.policy.policy_definition import (
    AffinityEntry,
    Carrier,
    CarrierEntry,
    ColorListEntry,
    CommunityAdditiveEntry,
    CommunityEntry,
    CommunityListEntry,
    DefaultAction,
    DefinitionWithSequencesCommonBase,
    DomainIDEntry,
    ExpandedCommunityListEntry,
    ExportToAction,
    GroupIDEntry,
    Match,
    MultiRegionRole,
    OMPTagEntry,
    OriginatorEntry,
    OriginEntry,
    OriginProtocol,
    PathType,
    PathTypeEntry,
    PolicyActionType,
    PolicyDefinitionBase,
    PolicyDefinitionSequenceBase,
    PreferenceEntry,
    PrefixListEntry,
    RegionEntry,
    RegionListEntry,
    RoleEntry,
    ServiceEntry,
    ServiceEntryValue,
    ServiceType,
    SiteEntry,
    SiteListEntry,
    TLOCActionEntry,
    TLOCActionType,
    TLOCEntry,
    TLOCEntryValue,
    TLOCListEntry,
    VPNListEntry,
    accept_action,
)

AnyControlPolicyRouteSequenceMatchEntry = Annotated[
    Union[
        ColorListEntry,
        CommunityListEntry,
        ExpandedCommunityListEntry,
        OMPTagEntry,
        OriginatorEntry,
        OriginEntry,
        PathTypeEntry,
        PreferenceEntry,
        PrefixListEntry,
        RegionEntry,
        RegionListEntry,
        RoleEntry,
        SiteEntry,
        SiteListEntry,
        SiteListEntry,
        TLOCEntry,
        TLOCListEntry,
        VPNListEntry,
    ],
    Field(discriminator="field"),
]

AnyControlPolicyTLOCSequenceMatchEntry = Annotated[
    Union[
        ColorListEntry,
        CommunityListEntry,
        CarrierEntry,
        ColorListEntry,
        DomainIDEntry,
        GroupIDEntry,
        OMPTagEntry,
        OriginatorEntry,
        PreferenceEntry,
        SiteEntry,
        SiteListEntry,
        RegionEntry,
        RegionListEntry,
        RoleEntry,
        TLOCListEntry,
        TLOCEntry,
    ],
    Field(discriminator="field"),
]

ControlPolicyRouteSequenceActions = Any  # TODO
ControlPolicyTLOCSequenceActions = Any  # TODO


class ControlPolicyHeader(PolicyDefinitionBase):
    type: Literal["control"] = "control"


class ControlPolicyRouteSequenceMatch(Match):
    entries: List[AnyControlPolicyRouteSequenceMatchEntry] = []


class ControlPolicyTLOCSequenceMatch(Match):
    entries: List[AnyControlPolicyTLOCSequenceMatchEntry] = []


class ControlPolicyRouteSequence(PolicyDefinitionSequenceBase):
    sequence_type: Literal["route"] = Field(
        default="route", serialization_alias="sequenceType", validation_alias="sequenceType"
    )
    base_action: PolicyActionType = Field(
        default="reject", serialization_alias="baseAction", validation_alias="baseAction"
    )
    match: ControlPolicyRouteSequenceMatch = ControlPolicyRouteSequenceMatch()
    actions: List[ControlPolicyRouteSequenceActions] = []
    model_config = ConfigDict(populate_by_name=True)

    def match_color_list(self, color_list_id: UUID) -> None:
        self._insert_match(ColorListEntry(ref=color_list_id))

    def match_community_list(self, community_list_id: UUID) -> None:
        self._insert_match(CommunityListEntry(ref=community_list_id))

    def match_expanded_community_list(self, expanded_community_list_id: UUID) -> None:
        self._insert_match(ExpandedCommunityListEntry(ref=expanded_community_list_id))

    def match_omp_tag(self, omp_tag: int) -> None:
        self._insert_match(OMPTagEntry(value=str(omp_tag)))

    def match_origin(self, origin: OriginProtocol) -> None:
        self._insert_match(OriginEntry(value=origin))

    def match_originator(self, originator: IPv4Address) -> None:
        self._insert_match(OriginatorEntry(value=originator))

    def match_path_type(self, path_type: PathType) -> None:
        self._insert_match(PathTypeEntry(value=path_type))

    def match_preference(self, preference: int) -> None:
        self._insert_match(PreferenceEntry(value=str(preference)))

    def match_region(self, region_id: int, role: Optional[MultiRegionRole] = None) -> None:
        self._insert_match(RegionEntry(value=str(region_id)))
        if role is not None:
            self._insert_match(RoleEntry(value=role))
        else:
            self._remove_match(RoleEntry)

    def match_region_list(self, region_list_id: UUID, role: Optional[MultiRegionRole] = None) -> None:
        self._insert_match(RegionListEntry(ref=region_list_id))
        if role is not None:
            self._insert_match(RoleEntry(value=role))
        else:
            self._remove_match(RoleEntry)

    def match_site(self, site: int) -> None:
        self._insert_match(SiteEntry(value=str(site)))

    def match_site_list(self, site_list_id: UUID) -> None:
        self._insert_match(SiteListEntry(ref=site_list_id))

    def match_tloc_list(self, tloc_list_id: UUID) -> None:
        self._insert_match(TLOCListEntry(ref=tloc_list_id))

    def match_tloc(self, ip: IPv4Address, color: TLOCColor, encap: EncapType) -> None:
        self._insert_match(TLOCEntry(value=TLOCEntryValue(ip=ip, color=color, encap=encap)))

    def match_vpn_list(self, vpn_list_id: UUID) -> None:
        self._insert_match(VPNListEntry(ref=vpn_list_id))

    def match_prefix_list(self, prefix_list_id: UUID) -> None:
        self._insert_match(PrefixListEntry(ref=prefix_list_id))

    @accept_action
    def associate_community_action(self, community: str, additive: bool = False) -> None:
        self._insert_action_in_set(CommunityEntry(value=community))
        if additive:
            self._insert_action_in_set(CommunityAdditiveEntry())
        else:
            self._remove_action_from_set(CommunityAdditiveEntry().field)

    @accept_action
    def associate_omp_tag_action(self, omp_tag: int) -> None:
        self._insert_action_in_set(OMPTagEntry(value=str(omp_tag)))

    @accept_action
    def associate_preference_action(self, preference: int) -> None:
        self._insert_action_in_set(PreferenceEntry(value=str(preference)))

    @overload
    def associate_service_action(self, service_type: ServiceType, vpn: int, *, tloc_list_id: UUID) -> None:
        ...

    @overload
    def associate_service_action(
        self, service_type: ServiceType, vpn: int, *, ip: IPv4Address, color: TLOCColor, encap: EncapType
    ) -> None:
        ...

    @accept_action
    def associate_service_action(
        self, service_type=ServiceType, vpn=int, *, tloc_list_id=None, ip=None, color=None, encap=None
    ) -> None:
        if tloc_list_id is None:
            tloc_entry = TLOCEntryValue(ip=ip, color=color, encap=encap)
            tloc_list_entry = None
        else:
            tloc_entry = None
            tloc_list_entry = TLOCListEntry(ref=tloc_list_id)
        service_value = ServiceEntryValue(type=service_type, vpn=str(vpn), tloc=tloc_entry, tloc_list=tloc_list_entry)
        self._insert_action_in_set(ServiceEntry(value=service_value))

    @accept_action
    def associate_tloc_action(self, tloc_action: TLOCActionType) -> None:
        self._insert_action_in_set(TLOCActionEntry(value=tloc_action))

    @accept_action
    def associate_affinity_action(self, affinity: int) -> None:
        self._insert_action_in_set(AffinityEntry(value=str(affinity)))

    @accept_action
    def associate_export_to_action(self, vpn_list_id: UUID) -> None:
        self._insert_action(ExportToAction(parameter=VPNListEntry(ref=vpn_list_id)))


class ControlPolicyTLOCSequence(PolicyDefinitionSequenceBase):
    sequence_type: Literal["tloc"] = Field(
        default="tloc", serialization_alias="sequenceType", validation_alias="sequenceType"
    )
    base_action: PolicyActionType = Field(
        default="reject", serialization_alias="baseAction", validation_alias="baseAction"
    )
    match: ControlPolicyTLOCSequenceMatch = ControlPolicyTLOCSequenceMatch()
    actions: List[ControlPolicyTLOCSequenceActions] = []
    model_config = ConfigDict(populate_by_name=True)

    def match_carrier(self, carrier: Carrier) -> None:
        self._insert_match(CarrierEntry(value=carrier))

    def match_color_list(self, color_list_id: UUID) -> None:
        self._insert_match(ColorListEntry(ref=color_list_id))

    def match_domain_id(self, domain_id: int) -> None:
        self._insert_match(DomainIDEntry(value=str(domain_id)))

    def match_group_id(self, group_id: int) -> None:
        self._insert_match(GroupIDEntry(value=str(group_id)))

    def match_omp_tag(self, omp_tag: int) -> None:
        self._insert_match(OMPTagEntry(value=str(omp_tag)))

    def match_originator(self, originator: IPv4Address) -> None:
        self._insert_match(OriginatorEntry(value=originator))

    def match_preference(self, preference: int) -> None:
        self._insert_match(PreferenceEntry(value=str(preference)))

    def match_site(self, site: int) -> None:
        self._insert_match(SiteEntry(value=str(site)))

    def match_site_list(self, site_list_id: UUID) -> None:
        self._insert_match(SiteListEntry(ref=site_list_id))

    def match_region(self, region_id: int, role: Optional[MultiRegionRole] = None) -> None:
        self._insert_match(RegionEntry(value=str(region_id)))
        if role is not None:
            self._insert_match(RoleEntry(value=role))
        else:
            self._remove_match(RoleEntry)

    def match_region_list(self, region_list_id: UUID, role: Optional[MultiRegionRole] = None) -> None:
        self._insert_match(RegionListEntry(ref=region_list_id))
        if role is not None:
            self._insert_match(RoleEntry(value=role))
        else:
            self._remove_match(RoleEntry)

    def match_tloc_list(self, tloc_list_id: UUID) -> None:
        self._insert_match(TLOCListEntry(ref=tloc_list_id))

    def match_tloc(self, ip: IPv4Address, color: TLOCColor, encap: EncapType) -> None:
        self._insert_match(TLOCEntry(value=TLOCEntryValue(ip=ip, color=color, encap=encap)))

    @accept_action
    def associate_omp_tag_action(self, omp_tag: int) -> None:
        self._insert_action_in_set(OMPTagEntry(value=str(omp_tag)))

    @accept_action
    def associate_preference_action(self, preference: int) -> None:
        self._insert_action_in_set(PreferenceEntry(value=str(preference)))

    @accept_action
    def associate_affinity_action(self, affinity: int) -> None:
        self._insert_action_in_set(AffinityEntry(value=str(affinity)))


AnyControlPolicySequence = Annotated[
    Union[ControlPolicyRouteSequence, ControlPolicyTLOCSequence],
    Field(discriminator="sequence_type"),
]


class ControlPolicy(ControlPolicyHeader, DefinitionWithSequencesCommonBase):
    sequences: List[AnyControlPolicySequence] = []
    default_action: DefaultAction = Field(
        default=DefaultAction(type="reject"),
        serialization_alias="defaultAction",
        validation_alias="defaultAction",
    )
    model_config = ConfigDict(populate_by_name=True)

    def add_route_sequence(
        self, name: str = "Route", base_action: PolicyActionType = "reject"
    ) -> ControlPolicyRouteSequence:
        seq = ControlPolicyRouteSequence(
            sequence_name=name,
            base_action=base_action,
            sequence_ip_type="ipv4",
        )
        self.add(seq)
        return seq

    def add_tloc_sequence(
        self, name: str = "TLOC", base_action: PolicyActionType = "reject"
    ) -> ControlPolicyTLOCSequence:
        seq = ControlPolicyTLOCSequence(
            sequence_name=name,
            base_action=base_action,
            sequence_ip_type="ipv4",
        )
        self.add(seq)
        return seq
